# Copyright (c) 2019 Zadara Storage, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for Zadara VPSA volume driver
"""
import copy
from unittest import mock

import requests
from six.moves.urllib import parse

from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.zadara import common
from cinder.volume.drivers.zadara import exception as zadara_exception
from cinder.volume.drivers.zadara import zadara


def check_access_key(func):
    """A decorator for all operations that needed an API before executing"""
    def wrap(self, *args, **kwargs):
        if not self._is_correct_access_key():
            return RUNTIME_VARS['bad_login']
        return func(self, *args, **kwargs)

    return wrap


DEFAULT_RUNTIME_VARS = {
    'status': 200,
    'user': 'test',
    'password': 'test_password',
    'access_key': '0123456789ABCDEF',
    'volumes': [],
    'servers': [],
    'controllers': [('active_ctrl', {'display-name': 'test_ctrl'})],
    'counter': 1000,

    "login": """
        {
         "response": {
                      "user": {
                               "updated-at": "2021-01-22",
                               "access-key": "%s",
                               "id": 1,
                               "created-at": "2021-01-22",
                               "email": "jsmith@example.com",
                               "username": "jsmith"
                              },
                      "status": 0
                     }
        }""",
    "good": """
         {
          "response": {
                       "status": 0
                      }
         }""",
    "bad_login": """
        {
         "response": {
                      "status": 5,
                      "status-msg": "Some message..."
                     }
        }""",
    "bad_volume": """
        {
         "response": {
                      "status": 10081,
                      "status-msg": "Virtual volume xxx should be found"
                     }
        }""",
    "fake_volume": """
        {
         "response": {
                      "volumes": [],
                      "status": 0,
                      "status-msg": "Virtual volume xxx doesn't exist"
                     }
        }""",
    "bad_server": """
        {
         "response": {
                      "status": 10086,
                      "status-msg": "Server xxx not found"
                     }
        }""",
    "server_created": """
        {
         "response": {
                      "server_name": "%s",
                      "status": 0
                     }
        }""",
}

RUNTIME_VARS = None


class FakeResponse(object):
    def __init__(self, method, url, params, body, headers, **kwargs):
        # kwargs include: verify, timeout
        self.method = method
        self.url = url
        self.body = body
        self.params = params
        self.headers = headers
        self.status = RUNTIME_VARS['status']

    @property
    def access_key(self):
        """Returns Response Access Key"""
        return self.headers["X-Access-Key"]

    def read(self):
        ops = {'POST': [('/api/users/login.json', self._login),
                        ('/api/volumes.json', self._create_volume),
                        ('/api/servers.json', self._create_server),
                        ('/api/servers/*/volumes.json', self._attach),
                        ('/api/volumes/*/detach.json', self._detach),
                        ('/api/volumes/*/expand.json', self._expand),
                        ('/api/volumes/*/rename.json', self._rename),
                        ('/api/consistency_groups/*/snapshots.json',
                         self._create_snapshot),
                        ('/api/snapshots/*/rename.json',
                         self._rename_snapshot),
                        ('/api/consistency_groups/*/clone.json',
                         self._create_clone)],
               'DELETE': [('/api/volumes/*', self._delete),
                          ('/api/snapshots/*', self._delete_snapshot)],
               'GET': [('/api/volumes.json?showonlyblock=YES',
                        self._list_volumes),
                       ('/api/volumes.json?display_name=*',
                           self._get_volume_by_name),
                       ('/api/pools/*.json', self._get_pool),
                       ('/api/vcontrollers.json', self._list_controllers),
                       ('/api/servers.json', self._list_servers),
                       ('/api/consistency_groups/*/snapshots.json',
                        self._list_vol_snapshots),
                       ('/api/snapshots.json', self._list_snapshots),
                       ('/api/volumes/*/servers.json',
                        self._list_vol_attachments)]
               }

        ops_list = ops[self.method]
        for (templ_url, func) in ops_list:
            if self._compare_url(self.url, templ_url):
                result = func()
                return result

    @staticmethod
    def _compare_url(url, template_url):
        items = url.split('/')
        titems = template_url.split('/')
        for (i, titem) in enumerate(titems):
            if '*' not in titem and titem != items[i]:
                return False
            if '?' in titem and titem.split('=')[0] != items[i].split('=')[0]:
                return False

        return True

    @staticmethod
    def _get_counter():
        cnt = RUNTIME_VARS['counter']
        RUNTIME_VARS['counter'] += 1
        return cnt

    def _login(self):
        params = self.body
        if (params['user'] == RUNTIME_VARS['user'] and
                params['password'] == RUNTIME_VARS['password']):
            return RUNTIME_VARS['login'] % RUNTIME_VARS['access_key']
        else:
            return RUNTIME_VARS['bad_login']

    def _is_correct_access_key(self):
        return self.access_key == RUNTIME_VARS['access_key']

    @check_access_key
    def _create_volume(self):
        params = self.body
        params['display-name'] = params['name']
        params['cg_name'] = params['name']
        params['snapshots'] = []
        params['server_ext_names'] = ''
        params['pool'] = 'pool-0001'
        params['provider_location'] = params['name']
        vpsa_vol = 'volume-%07d' % self._get_counter()
        RUNTIME_VARS['volumes'].append((vpsa_vol, params))
        return RUNTIME_VARS['good']

    @check_access_key
    def _create_server(self):
        params = self.body

        params['display-name'] = params['display_name']
        vpsa_srv = 'srv-%07d' % self._get_counter()
        RUNTIME_VARS['servers'].append((vpsa_srv, params))
        return RUNTIME_VARS['server_created'] % vpsa_srv

    @check_access_key
    def _attach(self):
        srv = self.url.split('/')[3]

        params = self.body

        vol = params['volume_name[]']

        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['name'] == vol:
                attachments = params['server_ext_names'].split(',')
                if srv in attachments:
                    # already attached - ok
                    return RUNTIME_VARS['good']
                else:
                    if not attachments[0]:
                        params['server_ext_names'] = srv
                    else:
                        params['server_ext_names'] += ',' + srv
                    return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _detach(self):
        params = self.body
        vol = self.url.split('/')[3]
        srv = params['server_name[]']

        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['name'] == vol:
                attachments = params['server_ext_names'].split(',')
                if srv not in attachments:
                    return RUNTIME_VARS['bad_server']
                else:
                    attachments.remove(srv)
                    params['server_ext_names'] = (','.join([str(elem)
                                                  for elem in attachments]))
                    return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _expand(self):
        params = self.body
        vol = self.url.split('/')[3]
        capacity = params['capacity']

        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['name'] == vol:
                params['capacity'] = capacity
                return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _rename(self):
        params = self.body
        vol = self.url.split('/')[3]

        for (vol_name, vol_params) in RUNTIME_VARS['volumes']:
            if vol_params['name'] == vol:
                vol_params['name'] = params['new_name']
                vol_params['display-name'] = params['new_name']
                vol_params['cg_name'] = params['new_name']
                return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _rename_snapshot(self):
        params = self.body
        vpsa_snapshot = self.url.split('/')[3]

        for (vol_name, vol_params) in RUNTIME_VARS['volumes']:
            for snapshot in vol_params['snapshots']:
                if vpsa_snapshot == snapshot:
                    vol_params['snapshots'].remove(snapshot)
                    vol_params['snapshots'].append(params['newname'])
                    return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _create_snapshot(self):
        params = self.body
        cg_name = self.url.split('/')[3]
        snap_name = params['display_name']

        for (vol_name, vol_params) in RUNTIME_VARS['volumes']:
            if vol_params['cg_name'] == cg_name:
                snapshots = vol_params['snapshots']
                if snap_name in snapshots:
                    # already attached
                    return RUNTIME_VARS['bad_volume']
                else:
                    snapshots.append(snap_name)
                    vol_params['has_snapshots'] = 'YES'
                    return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _delete_snapshot(self):
        snap = self.url.split('/')[3].split('.')[0]

        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if snap in params['snapshots']:
                params['snapshots'].remove(snap)
                return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    @check_access_key
    def _create_clone(self):
        params = self.body
        params['display-name'] = params['name']
        params['cg_name'] = params['name']
        params['capacity'] = 1
        params['snapshots'] = []
        params['server_ext_names'] = ''
        params['pool'] = 'pool-0001'
        vpsa_vol = 'volume-%07d' % self._get_counter()
        RUNTIME_VARS['volumes'].append((vpsa_vol, params))
        return RUNTIME_VARS['good']

    def _delete(self):
        vol = self.url.split('/')[3].split('.')[0]

        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['name'] == vol:
                if params['server_ext_names']:
                    # there are attachments - should be volume busy error
                    return RUNTIME_VARS['bad_volume']
                else:
                    RUNTIME_VARS['volumes'].remove((vol_name, params))
                    return RUNTIME_VARS['good']

        return RUNTIME_VARS['bad_volume']

    def _generate_list_resp(self, null_body, body, lst, vol):
        resp = ''
        for (obj, params) in lst:
            if vol:
                resp += body % (obj,
                                params['display-name'],
                                params['cg_name'],
                                params['capacity'],
                                params['pool'])
            else:
                resp += body % (obj, params['display-name'])
        if resp:
            return resp
        else:
            return null_body

    def _list_volumes(self):
        null_body = """
        {
         "response": {
                      "volumes": [
                                 ],
                      "status": 0
                     }
        }"""
        body = """
        {
         "response": {
                      "volumes": %s,
                      "status": 0
                     }
        }"""

        volume_obj = """
                     {
                      "name": "%s",
                      "display_name": "%s",
                      "cg_name": "%s",
                      "status": "%s",
                      "virtual_capacity": %d,
                      "pool_name": "%s",
                      "allocated-capacity": 1,
                      "raid-group-name": "r5",
                      "cache": "write-through",
                      "created-at": "2021-01-22",
                      "modified-at": "2021-01-22",
                      "has_snapshots": "%s"
                     }
                     """
        if len(RUNTIME_VARS['volumes']) == 0:
            return null_body
        resp = ''
        volume_list = ''
        count = 0
        for (vol_name, params) in RUNTIME_VARS['volumes']:
            vol_status = (params.get('status') if params.get('status')
                          else 'Available')
            has_snapshots = 'YES' if params.get('has_snapshots') else 'NO'
            volume_dict = volume_obj % (params['name'],
                                        params['display-name'],
                                        params['cg_name'],
                                        vol_status,
                                        params['capacity'],
                                        params['pool'],
                                        has_snapshots)
            if count == 0:
                volume_list += volume_dict
                count += 1
            elif count != len(RUNTIME_VARS['volumes']):
                volume_list = volume_list + ',' + volume_dict
                count += 1
        if volume_list:
            volume_list = '[' + volume_list + ']'
            resp = body % volume_list
            return resp

        return RUNTIME_VARS['bad_volume']

    def _get_volume_by_name(self):
        volume_name = self.url.split('=')[1]
        body = """
        {
         "response": {
                      "volumes": [
                                  {
                                   "name": "%s",
                                   "display_name": "%s",
                                   "cg_name": "%s",
                                   "provider_location": "%s",
                                   "status": "%s",
                                   "virtual_capacity": %d,
                                   "pool_name": "%s",
                                   "allocated-capacity": 1,
                                   "raid-group-name": "r5",
                                   "cache": "write-through",
                                   "created-at": "2021-01-22",
                                   "modified-at": "2021-01-22",
                                   "has_snapshots": "%s",
                                   "server_ext_names": "%s"
                                  }
                                 ],
                      "status": 0
                     }
        }"""
        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['name'] == volume_name:
                vol_status = (params.get('status') if params.get('status')
                              else 'Available')
                has_snapshots = 'YES' if params.get('has_snapshots') else 'NO'
                resp = body % (volume_name, params['display-name'],
                               params['cg_name'],
                               params['cg_name'],
                               vol_status,
                               params['capacity'],
                               params['pool'],
                               has_snapshots,
                               params['server_ext_names'])
                return resp

        return RUNTIME_VARS['fake_volume']

    def _list_controllers(self):
        null_body = """
        {
         "response": {
                      "vcontrollers": [
                                      ],
                      "status": 0
                     }
        }"""
        body = """
        {
         "response": {
                      "vcontrollers": [
                                       {
                                        "name": "%s",
                                        "display-name": "%s",
                                        "state": "active",
                                        "target":
                                        "iqn.2011-04.zadarastorage:vsa-xxx:1",
                                        "iscsi_ip": "1.1.1.1",
                                        "iscsi_ipv6": "",
                                        "mgmt-ip": "1.1.1.1",
                                        "software-ver": "0.0.09-05.1--77.7",
                                        "heartbeat1": "ok",
                                        "heartbeat2": "ok",
                                        "vpsa_chap_user": "test_chap_user",
                                        "vpsa_chap_secret": "test_chap_secret"
                                       }
                                      ],
                      "status": 0
                     }
        }"""
        return self._generate_list_resp(null_body,
                                        body,
                                        RUNTIME_VARS['controllers'],
                                        False)

    def _get_pool(self):
        response = """
        {
         "response": {
                      "pool": {
                               "name": "pool-0001",
                               "capacity": 100,
                               "available_capacity": 99,
                               "provisioned_capacity": 1
                              },
                      "status": 0
                     }
        }"""
        return response

    def _list_servers(self):
        null_body = """
        {
         "response": {
                      "servers": [
                                 ],
                      "status": 0
                     }
        }"""
        body = """
        {
         "response": {
                      "servers": %s,
                      "status": 0
                     }
        }"""

        server_obj = """
                     {
                      "name": "%s",
                      "display_name": "%s",
                      "iqn": "%s",
                      "target":
                      "iqn.2011-04.zadarastorage:vsa-xxx:1",
                      "lun": 0
                     }
                     """

        resp = ''
        server_list = ''
        count = 0
        for (obj, params) in RUNTIME_VARS['servers']:
            server_dict = server_obj % (obj,
                                        params['display_name'],
                                        params['iqn'])
            if count == 0:
                server_list += server_dict
                count += 1
            elif count != len(RUNTIME_VARS['servers']):
                server_list = server_list + ',' + server_dict
                count += 1
        server_list = '[' + server_list + ']'
        resp = body % server_list
        if resp:
            return resp
        else:
            return null_body

    def _list_snapshots(self):
        null_body = """
        {
         "response": {
                      "snapshots": [
                                 ],
                      "status": 0
                     }
        }"""
        body = """
        {
         "response": {
                      "snapshots": %s,
                      "status": 0
                     }
        }"""

        snapshot_obj = """
                       {
                        "name": "%s",
                        "display_name": "%s",
                        "volume_display_name": "%s",
                        "volume_capacity_mb": %d,
                        "volume_ext_name": "%s",
                        "cg_name": "%s",
                        "pool_name": "pool-0001"
                     }
                     """

        resp = ''
        snapshot_list = ''
        count = 0
        for (obj, params) in RUNTIME_VARS['volumes']:
            snapshots = params['snapshots']
            if len(snapshots) == 0:
                continue
            for snapshot in snapshots:
                snapshot_dict = snapshot_obj % (snapshot, snapshot,
                                                params['provider_location'],
                                                params['capacity'] * 1024,
                                                params['display-name'],
                                                params['cg_name'])
                if count == 0:
                    snapshot_list += snapshot_dict
                    count += 1
                else:
                    snapshot_list = snapshot_list + ',' + snapshot_dict
                    count += 1
        snapshot_list = '[' + snapshot_list + ']'
        resp = body % snapshot_list
        if resp:
            return resp
        else:
            return null_body

    def _get_server_obj(self, name):
        for (srv_name, params) in RUNTIME_VARS['servers']:
            if srv_name == name:
                return params

    def _list_vol_attachments(self):
        vol = self.url.split('/')[3]

        null_body = """
        {
         "response": {
                      "servers": [
                                 ],
                      "status": 0
                     }
        }"""
        body = """
        {
         "response": {
                      "servers": %s,
                      "status": 0
                     }
        }"""

        server_obj = """
                     {
                      "name": "%s",
                      "display_name": "%s",
                      "iqn": "%s",
                      "target":
                      "iqn.2011-04.zadarastorage:vsa-xxx:1",
                      "lun": 0
                     }
                     """
        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['name'] == vol:
                attachments = params['server_ext_names'].split(',')
                if not attachments[0]:
                    return null_body
                resp = ''
                server_list = ''
                count = 0
                for server in attachments:
                    srv_params = self._get_server_obj(server)
                    server_dict = (server_obj % (server,
                                   srv_params['display_name'],
                                   srv_params['iqn']))
                    if count == 0:
                        server_list += server_dict
                        count += 1
                    elif count != len(attachments):
                        server_list = server_list + ',' + server_dict
                        count += 1
                server_list = '[' + server_list + ']'
                resp = body % server_list
                return resp

        return RUNTIME_VARS['bad_volume']

    def _list_vol_snapshots(self):
        cg_name = self.url.split('/')[3]

        null_body = """
        {
         "response": {
                      "snapshots": [
                                   ],
                      "status": 0
                     }
        }"""

        body = """
        {
         "response": {
                      "snapshots": %s,
                      "status": 0
                     }
        }"""

        snapshot_obj = """
                       {
                        "name": "%s",
                        "display_name": "%s",
                        "cg_name": "%s",
                        "pool_name": "pool-0001"
                       }
                       """
        for (vol_name, params) in RUNTIME_VARS['volumes']:
            if params['cg_name'] == cg_name:
                snapshots = params['snapshots']
                if len(snapshots) == 0:
                    return null_body
                resp = ''
                snapshot_list = ''
                count = 0

                for snapshot in snapshots:
                    snapshot_dict = snapshot_obj % (snapshot, snapshot,
                                                    cg_name)
                    if count == 0:
                        snapshot_list += snapshot_dict
                        count += 1
                    elif count != len(snapshots):
                        snapshot_list = snapshot_list + ',' + snapshot_dict
                        count += 1
                snapshot_list = '[' + snapshot_list + ']'
                resp = body % snapshot_list
                return resp

        return RUNTIME_VARS['bad_volume']


class FakeRequests(object):
    """A fake requests for zadara volume driver tests."""
    def __init__(self, method, api_url, params=None, data=None,
                 headers=None, **kwargs):
        apiurl_items = parse.urlparse(api_url)
        if apiurl_items.query:
            url = apiurl_items.path + '?' + apiurl_items.query
        else:
            url = apiurl_items.path
        res = FakeResponse(method, url, params, data, headers, **kwargs)
        self.content = res.read()
        self.status_code = res.status


class ZadaraVPSADriverTestCase(test.TestCase):

    def __init__(self, *args, **kwargs):
        super(ZadaraVPSADriverTestCase, self).__init__(*args, **kwargs)

        self.configuration = None
        self.driver = None

    """Test case for Zadara VPSA volume driver."""
    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def setUp(self):
        super(ZadaraVPSADriverTestCase, self).setUp()

        global RUNTIME_VARS
        RUNTIME_VARS = copy.deepcopy(DEFAULT_RUNTIME_VARS)
        self.configuration = mock.Mock(conf.Configuration(None))
        self.configuration.append_config_values(common.zadara_opts)
        self.configuration.reserved_percentage = 10
        self.configuration.zadara_use_iser = True
        self.configuration.zadara_vpsa_host = '192.168.5.5'
        self.configuration.zadara_vpsa_port = '80'
        self.configuration.zadara_user = 'test'
        self.configuration.zadara_password = 'test_password'
        self.configuration.zadara_access_key = '0123456789ABCDEF'
        self.configuration.zadara_vpsa_poolname = 'pool-0001'
        self.configuration.zadara_vol_encrypt = False
        self.configuration.zadara_vol_name_template = 'OS_%s'
        self.configuration.zadara_vpsa_use_ssl = False
        self.configuration.zadara_ssl_cert_verify = False
        self.configuration.driver_ssl_cert_path = '/path/to/cert'
        self.configuration.zadara_default_snap_policy = False
        self.configuration.zadara_gen3_vol_compress = False
        self.configuration.zadara_gen3_vol_dedupe = False
        self.driver = (zadara.ZadaraVPSAISCSIDriver(
                       configuration=self.configuration))
        self.driver.do_setup(None)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_create_destroy(self):
        """Create/Delete volume."""
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.driver.create_volume(volume)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_create_destroy_multiple(self):
        """Create/Delete multiple volumes."""
        vol1_args = {'display_name': 'test_volume_01', 'size': 1,
                     'id': fake.VOLUME_ID}
        vol2_args = {'display_name': 'test_volume_02', 'size': 2,
                     'id': fake.VOLUME2_ID}
        vol3_args = {'display_name': 'test_volume_03', 'size': 3,
                     'id': fake.VOLUME3_ID}
        vol4_args = {'display_name': 'test_volume_04', 'size': 4,
                     'id': fake.VOLUME4_ID}
        volume1 = fake_volume.fake_volume_obj(None, **vol1_args)
        volume2 = fake_volume.fake_volume_obj(None, **vol2_args)
        volume3 = fake_volume.fake_volume_obj(None, **vol3_args)
        volume4 = fake_volume.fake_volume_obj(None, **vol4_args)

        self.driver.create_volume(volume1)
        self.driver.create_volume(volume2)
        self.driver.create_volume(volume3)
        self.driver.delete_volume(volume1)
        self.driver.delete_volume(volume2)
        self.driver.delete_volume(volume3)
        self.driver.delete_volume(volume4)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_destroy_non_existent(self):
        """Delete non-existent volume."""
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_empty_apis(self):
        """Test empty func (for coverage only)."""
        context = None
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.driver.create_export(context, volume)
        self.driver.ensure_export(context, volume)
        self.driver.remove_export(context, volume)
        self.assertRaises(NotImplementedError,
                          self.driver.local_path,
                          None)
        self.driver.check_for_setup_error()

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_volume_attach_detach(self):
        """Test volume attachment and detach."""
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        connector = dict(initiator='test_iqn.1')
        self.driver.create_volume(volume)
        props = self.driver.initialize_connection(volume, connector)
        self.assertEqual('iser', props['driver_volume_type'])
        data = props['data']
        self.assertEqual('1.1.1.1:3260', data['target_portal'])
        self.assertEqual('iqn.2011-04.zadarastorage:vsa-xxx:1',
                         data['target_iqn'])
        self.assertEqual(int('0'), data['target_lun'])
        self.assertEqual(volume['id'], data['volume_id'])
        self.assertEqual('CHAP', data['auth_method'])
        self.assertEqual('test_chap_user', data['auth_username'])
        self.assertEqual('test_chap_secret', data['auth_password'])
        self.driver.terminate_connection(volume, connector)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_wrong_attach_params(self):
        """Test different wrong attach scenarios."""
        vol1_args = {'display_name': 'test_volume_01', 'size': 1,
                     'id': fake.VOLUME_ID}
        volume1 = fake_volume.fake_volume_obj(None, **vol1_args)
        connector1 = dict(initiator='test_iqn.1')
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.initialize_connection,
                          volume1, connector1)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_wrong_detach_params(self):
        """Test different wrong detachment scenarios."""
        vol1_args = {'display_name': 'test_volume_01', 'size': 1,
                     'id': fake.VOLUME_ID}
        volume1 = fake_volume.fake_volume_obj(None, **vol1_args)
        # Volume is not created.
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.terminate_connection,
                          volume1, None)

        self.driver.create_volume(volume1)
        connector1 = dict(initiator='test_iqn.1')
        # Server is not found. Volume is found
        self.assertRaises(zadara_exception.ZadaraServerNotFound,
                          self.driver.terminate_connection,
                          volume1, connector1)

        vol2_args = {'display_name': 'test_volume_02', 'size': 1,
                     'id': fake.VOLUME2_ID}
        vol3_args = {'display_name': 'test_volume_03', 'size': 1,
                     'id': fake.VOLUME3_ID}
        volume2 = fake_volume.fake_volume_obj(None, **vol2_args)
        volume3 = fake_volume.fake_volume_obj(None, **vol3_args)
        connector2 = dict(initiator='test_iqn.2')
        connector3 = dict(initiator='test_iqn.3')
        self.driver.create_volume(volume2)
        self.driver.initialize_connection(volume1, connector1)
        self.driver.initialize_connection(volume2, connector2)
        # volume is found. Server not found
        self.assertRaises(zadara_exception.ZadaraServerNotFound,
                          self.driver.terminate_connection,
                          volume1, connector3)
        # Server is found. volume not found
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.terminate_connection,
                          volume3, connector1)
        # Server and volume exits but not attached
        self.assertRaises(common.exception.FailedCmdWithDump,
                          self.driver.terminate_connection,
                          volume1, connector2)

        self.driver.terminate_connection(volume1, connector1)
        self.driver.terminate_connection(volume2, connector2)

    @mock.patch.object(requests.Session, 'request')
    def test_ssl_use(self, request):
        """Coverage test for SSL connection."""
        self.configuration.zadara_ssl_cert_verify = True
        self.configuration.zadara_vpsa_use_ssl = True
        self.configuration.driver_ssl_cert_path = '/path/to/cert'

        fake_request_ctrls = FakeRequests("GET", "/api/vcontrollers.json")
        raw_controllers = fake_request_ctrls.content
        good_response = mock.MagicMock()
        good_response.status_code = RUNTIME_VARS['status']
        good_response.content = raw_controllers

        def request_verify_cert(*args, **kwargs):
            self.assertEqual(kwargs['verify'], '/path/to/cert')
            return good_response

        request.side_effect = request_verify_cert
        self.driver.do_setup(None)

    @mock.patch.object(requests.Session, 'request')
    def test_wrong_access_key(self, request):
        """Wrong Access Key"""
        fake_ak = 'FAKEACCESSKEY'
        self.configuration.zadara_access_key = fake_ak

        bad_response = mock.MagicMock()
        bad_response.status_code = RUNTIME_VARS['status']
        bad_response.content = RUNTIME_VARS['bad_login']

        def request_verify_access_key(*args, **kwargs):
            # Checks if the fake access_key was sent to driver
            token = kwargs['headers']['X-Access-Key']
            self.assertEqual(token, fake_ak, "access_key wasn't delivered")
            return bad_response

        request.side_effect = request_verify_access_key
        # when access key is invalid, driver will raise
        # ZadaraInvalidAccessKey exception
        self.assertRaises(zadara_exception.ZadaraCinderInvalidAccessKey,
                          self.driver.do_setup,
                          None)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_bad_http_response(self):
        """Coverage test for non-good HTTP response."""
        RUNTIME_VARS['status'] = 400

        vol_args = {'display_name': 'test_volume_03', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.assertRaises(exception.BadHTTPResponseStatus,
                          self.driver.create_volume, volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_terminate_connection_force_detach(self):
        """Test terminate connection for os-force_detach """
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        connector = dict(initiator='test_iqn.1')

        self.driver.create_volume(volume)
        self.driver.initialize_connection(volume, connector)

        # connector is None - force detach - detach all mappings
        self.driver.terminate_connection(volume, None)

        self.assertRaises(common.exception.FailedCmdWithDump,
                          self.driver.terminate_connection,
                          volume, connector)

        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_delete_without_detach(self):
        """Test volume deletion without detach."""

        vol1_args = {'display_name': 'test_volume_01', 'size': 1,
                     'id': fake.VOLUME_ID}
        volume1 = fake_volume.fake_volume_obj(None, **vol1_args)
        connector1 = dict(initiator='test_iqn.1')
        connector2 = dict(initiator='test_iqn.2')
        connector3 = dict(initiator='test_iqn.3')
        self.driver.create_volume(volume1)
        self.driver.initialize_connection(volume1, connector1)
        self.driver.initialize_connection(volume1, connector2)
        self.driver.initialize_connection(volume1, connector3)
        self.driver.delete_volume(volume1)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_no_active_ctrl(self):

        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        connector = dict(initiator='test_iqn.1')
        self.driver.create_volume(volume)

        RUNTIME_VARS['controllers'] = []
        self.assertRaises(zadara_exception.ZadaraVPSANoActiveController,
                          self.driver.initialize_connection,
                          volume, connector)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_create_destroy_snapshot(self):
        """Create/Delete snapshot test."""
        wrong_vol_args = {'display_name': 'wrong_vol_01', 'size': 1,
                          'id': fake.VOLUME2_ID}
        wrong_volume = fake_volume.fake_volume_obj(None, **wrong_vol_args)
        wrong_snap_args = {'display_name': 'snap_01', 'volume': wrong_volume}
        wrong_snapshot = fake_snapshot.fake_snapshot_obj(None,
                                                         **wrong_snap_args)
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_snapshot,
                          wrong_snapshot)

        # Create cinder volume and snapshot
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        snap_args = {'display_name': 'test_snap_01', 'id': fake.SNAPSHOT_ID,
                     'volume': volume}
        snapshot = fake_snapshot.fake_snapshot_obj(None, **snap_args)
        self.driver.create_volume(volume)
        self.driver.create_snapshot(snapshot)

        # Deleted should succeed for missing volume
        self.driver.delete_snapshot(wrong_snapshot)

        # Deleted should succeed for missing snap
        fake_snap_args = {'display_name': 'test_snap_02',
                          'id': fake.SNAPSHOT2_ID, 'volume': volume}
        fake_snap = fake_snapshot.fake_snapshot_obj(None, **fake_snap_args)
        self.driver.delete_snapshot(fake_snap)

        self.driver.delete_snapshot(snapshot)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_expand_volume(self):
        """Expand volume test."""
        vol_args = {'display_name': 'test_volume_01', 'id': fake.VOLUME_ID,
                    'size': 10}
        vol2_args = {'display_name': 'test_volume_02', 'id': fake.VOLUME2_ID,
                     'size': 10}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        volume2 = fake_volume.fake_volume_obj(None, **vol2_args)

        self.driver.create_volume(volume)

        self.assertRaises(exception.VolumeDriverException,
                          self.driver.extend_volume,
                          volume2, 15)
        self.assertRaises(exception.InvalidInput,
                          self.driver.extend_volume,
                          volume, 5)

        self.driver.extend_volume(volume, 15)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_create_destroy_clones(self):
        """Create/Delete clones test."""
        vol1_args = {'display_name': 'test_volume_01', 'id': fake.VOLUME_ID,
                     'size': 1}
        vol2_args = {'display_name': 'test_volume_02', 'id': fake.VOLUME2_ID,
                     'size': 2}
        vol3_args = {'display_name': 'test_volume_03', 'id': fake.VOLUME3_ID,
                     'size': 1}
        volume1 = fake_volume.fake_volume_obj(None, **vol1_args)
        volume2 = fake_volume.fake_volume_obj(None, **vol2_args)
        volume3 = fake_volume.fake_volume_obj(None, **vol3_args)

        snap_args = {'display_name': 'test_snap_01',
                     'id': fake.SNAPSHOT_ID, 'volume': volume1}
        snapshot = fake_snapshot.fake_snapshot_obj(None, **snap_args)
        self.driver.create_volume(volume1)
        self.driver.create_snapshot(snapshot)

        # Test invalid vol reference
        wrong_vol_args = {'display_name': 'wrong_volume_01',
                          'id': fake.VOLUME4_ID, 'size': 1}
        wrong_volume = fake_volume.fake_volume_obj(None, **wrong_vol_args)
        wrong_snap_args = {'display_name': 'test_wrong_snap',
                           'id': fake.SNAPSHOT2_ID, 'volume': wrong_volume}
        wrong_snapshot = fake_snapshot.fake_snapshot_obj(None,
                                                         **wrong_snap_args)
        self.assertRaises(exception.SnapshotNotFound,
                          self.driver.create_volume_from_snapshot,
                          wrong_volume,
                          wrong_snapshot)

        wrong_snap_args = {'display_name': 'test_wrong_snap',
                           'id': fake.SNAPSHOT3_ID, 'volume': volume1}
        wrong_snapshot = fake_snapshot.fake_snapshot_obj(None,
                                                         **wrong_snap_args)
        # Test invalid snap reference
        self.assertRaises(exception.SnapshotNotFound,
                          self.driver.create_volume_from_snapshot,
                          volume1,
                          wrong_snapshot)
        # Test invalid src_vref for volume clone
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_cloned_volume,
                          volume3, volume2)
        self.driver.create_volume_from_snapshot(volume2, snapshot)
        self.driver.create_cloned_volume(volume3, volume1)
        self.driver.delete_volume(volume3)
        self.driver.delete_volume(volume2)
        self.driver.delete_snapshot(snapshot)
        self.driver.delete_volume(volume1)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_get_volume_stats(self):
        """Get stats test."""
        self.configuration.safe_get.return_value = 'ZadaraVPSAISCSIDriver'
        data = self.driver.get_volume_stats(True)
        self.assertEqual('Zadara Storage', data['vendor_name'])
        self.assertEqual(100, data['total_capacity_gb'])
        self.assertEqual(99, data['free_capacity_gb'])
        self.assertEqual({'total_capacity_gb': 100,
                          'free_capacity_gb': 99,
                          'multiattach': True,
                          'reserved_percentage':
                          self.configuration.reserved_percentage,
                          'QoS_support': False,
                          'vendor_name': 'Zadara Storage',
                          'driver_version': self.driver.VERSION,
                          'storage_protocol': 'iSER',
                          'volume_backend_name': 'ZadaraVPSAISCSIDriver'},
                         data)

    def create_vpsa_backend_volume(self, vol_id, vol_name, vol_size,
                                   vol_status, has_snapshots):
        vol_params = {}
        vol_params['id'] = vol_id
        vol_params['name'] = vol_name
        vol_params['display-name'] = vol_name
        vol_params['cg_name'] = vol_name
        vol_params['provider_location'] = vol_name
        vol_params['status'] = vol_status
        vol_params['capacity'] = vol_size
        vol_params['pool'] = 'pool-0001'
        vol_params['has_snapshots'] = has_snapshots
        vol_params['server_ext_names'] = ''
        vol_params['snapshots'] = []
        volname = 'fake-volume'
        vpsa_volume = (volname, vol_params)
        RUNTIME_VARS['volumes'].append(vpsa_volume)
        return vpsa_volume

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_manage_existing_volume(self):
        vol_args = {'id': fake.VOLUME_ID,
                    'display_name': 'manage-name',
                    'size': 1}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        vpsa_volume = self.create_vpsa_backend_volume('fake_id',
                                                      'fake_name', 1,
                                                      'Available', 'NO')
        # Check the failure with an empty reference for volume
        identifier = {}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing,
                          volume, identifier)

        # Check the failure with an invalid reference for volume
        identifier['name'] = 'fake_identifier'
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing,
                          volume, identifier)

        identifier['name'] = 'fake_name'
        self.driver.manage_existing(volume, identifier)
        # Check the new volume renamed accordingly
        self.assertEqual(vpsa_volume[1]['display-name'],
                         'OS_%s' % volume['name'])
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_manage_existing_snapshot(self):
        vol_args = {'display_name': 'fake_name', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.driver.create_volume(volume)

        # Create a backend snapshot that will be managed by cinder volume
        (vol_name, vol_params) = RUNTIME_VARS['volumes'][0]
        vol_params['snapshots'].append('fakesnapname')

        # Check the failure with wrong volume for snapshot
        wrong_vol_args = {'display_name': 'wrong_volume_01',
                          'size': 1, 'id': fake.VOLUME2_ID}
        wrong_volume = fake_volume.fake_volume_obj(None, **wrong_vol_args)
        wrong_snap_args = {'display_name': 'snap_01', 'volume': wrong_volume}
        wrong_snapshot = fake_snapshot.fake_snapshot_obj(None,
                                                         **wrong_snap_args)
        identifier = {}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          wrong_snapshot, identifier)

        identifier['name'] = 'fake_identifier'
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          wrong_snapshot, identifier)

        # Check the failure with wrong identifier for the snapshot
        snap_args = {'display_name': 'manage_snapname',
                     'id': fake.SNAPSHOT_ID, 'volume': volume}
        snapshot = fake_snapshot.fake_snapshot_obj(None, **snap_args)
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          snapshot, identifier)

        identifier['name'] = 'fakesnapname'
        self.driver.manage_existing_snapshot(snapshot, identifier)
        # Check that the backend snapshot has been renamed
        (vol_name, vol_params) = RUNTIME_VARS['volumes'][0]
        self.assertEqual(vol_params['snapshots'][0], snapshot['name'])
        self.driver.delete_snapshot(snapshot)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_get_manageable_volumes(self):
        vpsa_volume1 = self.create_vpsa_backend_volume('manage_vol_id1',
                                                       'manage_vol1', 1,
                                                       'Available', 'NO')
        vpsa_volume2 = self.create_vpsa_backend_volume('manage_vol_id2',
                                                       'manage_vol2', 2,
                                                       'Available', 'NO')

        cinder_vol1_args = {'display_name': 'fake-volume1',
                            'size': 3, 'id': fake.VOLUME_ID}
        cinder_vol2_args = {'display_name': 'fake-volume2',
                            'size': 4, 'id': fake.VOLUME2_ID}
        cinder_vol1 = fake_volume.fake_volume_obj(None, **cinder_vol1_args)
        cinder_vol2 = fake_volume.fake_volume_obj(None, **cinder_vol2_args)
        self.driver.create_volume(cinder_vol1)
        self.driver.create_volume(cinder_vol2)

        cinder_vols = [cinder_vol1, cinder_vol2]
        manageable_vols = (self.driver.get_manageable_volumes(
                           cinder_vols, None, 10, 0, ['size'], ['asc']))
        # Check the volumes are returned in the sorted order
        self.assertEqual(len(manageable_vols), 4)
        self.assertGreater(manageable_vols[1]['size'],
                           manageable_vols[0]['size'])
        self.assertGreater(manageable_vols[3]['size'],
                           manageable_vols[2]['size'])
        self.driver.delete_volume(cinder_vol1)
        self.driver.delete_volume(cinder_vol2)

        # Try to manage the volume and delete it
        vol1_args = {'display_name': 'manage-name1',
                     'size': 1, 'id': fake.VOLUME3_ID}
        volume1 = fake_volume.fake_volume_obj(None, **vol1_args)
        identifier = {'name': 'manage_vol1'}
        self.driver.manage_existing(volume1, identifier)
        self.assertEqual(vpsa_volume1[1]['display-name'],
                         'OS_%s' % volume1['name'])
        self.driver.delete_volume(volume1)

        # Manage and delete the volume
        vol2_args = {'display_name': 'manage-name2',
                     'size': 1, 'id': fake.VOLUME4_ID}
        volume2 = fake_volume.fake_volume_obj(None, **vol2_args)
        identifier = {'name': 'manage_vol2'}
        self.driver.manage_existing(volume2, identifier)
        self.assertEqual(vpsa_volume2[1]['display-name'],
                         'OS_%s' % volume2['name'])
        self.driver.delete_volume(volume2)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_get_manageable_snapshots(self):
        # Create a cinder volume and a snapshot
        vol_args = {'display_name': 'test_volume_01', 'size': 1,
                    'id': fake.VOLUME_ID}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        snap_args = {'display_name': 'test_snap_01',
                     'id': fake.SNAPSHOT_ID, 'volume': volume}
        snapshot = fake_snapshot.fake_snapshot_obj(None, **snap_args)
        self.driver.create_volume(volume)
        self.driver.create_snapshot(snapshot)

        # Create backend snapshots for the volume
        vpsa_volume = self.create_vpsa_backend_volume('manage_vol_id',
                                                      'manage_vol', 1,
                                                      'Available', 'YES')
        snapshot1 = {'name': 'manage_snap_01',
                     'volume_name': vpsa_volume[1]['name'],
                     'provider_location': 'manage_snap_01'}
        snapshot2 = {'name': 'manage_snap_02',
                     'volume_name': vpsa_volume[1]['name'],
                     'provider_location': 'manage_snap_02'}
        vpsa_volume[1]['snapshots'].append(snapshot1['name'])
        vpsa_volume[1]['snapshots'].append(snapshot2['name'])

        cinder_snapshots = [snapshot]
        manageable_snapshots = (self.driver.get_manageable_snapshots(
            cinder_snapshots, None, 10, 0, ['reference'], ['asc']))
        # Check the returned manageable snapshot names
        self.assertEqual(snapshot1['name'],
                         manageable_snapshots[0]['reference']['name'])
        self.assertEqual(snapshot2['name'],
                         manageable_snapshots[1]['reference']['name'])

        # Verify the safety of the snapshots to manage
        self.assertEqual(manageable_snapshots[0]['safe_to_manage'], True)
        self.assertEqual(manageable_snapshots[1]['safe_to_manage'], True)

        # Verify the refernce of the source volume of the snapshots
        source_vol = manageable_snapshots[0]['source_reference']
        self.assertEqual(vpsa_volume[1]['name'], source_vol['name'])
        source_vol = manageable_snapshots[1]['source_reference']
        self.assertEqual(vpsa_volume[1]['name'], source_vol['name'])
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_manage_existing_volume_get_size(self):
        vol_args = {'display_name': 'fake_name', 'id': fake.VOLUME_ID,
                    'size': 1}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.driver.create_volume(volume)

        # Check the failure with empty reference of the volume
        identifier = {}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_get_size,
                          volume, identifier)

        # Check the failure with invalid volume reference
        identifier = {'name': 'fake_identifiter'}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_get_size,
                          volume, identifier)

        # Verify the volume size
        identifier = {'name': 'OS_volume-%s' % volume['id']}
        vol_size = self.driver.manage_existing_get_size(volume, identifier)
        self.assertEqual(vol_size, volume.size)
        self.driver.delete_volume(volume)

    @mock.patch.object(requests.Session, 'request', FakeRequests)
    def test_manage_existing_snapshot_get_size(self):
        # Create a cinder volume and a snapshot
        vol_args = {'display_name': 'fake_name', 'id': fake.VOLUME_ID,
                    'size': 1}
        volume = fake_volume.fake_volume_obj(None, **vol_args)
        self.driver.create_volume(volume)
        snap_args = {'display_name': 'fake_snap',
                     'id': fake.SNAPSHOT_ID, 'volume': volume}
        snapshot = fake_snapshot.fake_snapshot_obj(None, **snap_args)
        self.driver.create_snapshot(snapshot)

        # Check with the wrong volume of the snapshot
        wrong_vol_args = {'display_name': 'wrong_volume_01',
                          'size': 1, 'id': fake.VOLUME2_ID}
        wrong_volume = fake_volume.fake_volume_obj(None, **wrong_vol_args)
        wrong_snap_args = {'display_name': 'wrong_snap',
                           'volume': wrong_volume}
        wrong_snapshot = fake_snapshot.fake_snapshot_obj(None,
                                                         **wrong_snap_args)
        identifier = {}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot_get_size,
                          wrong_snapshot, identifier)

        identifier = {'name': 'fake_identifiter'}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot_get_size,
                          wrong_snapshot, identifier)

        # Check with the invalid reference of the snapshot
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot_get_size,
                          snapshot, identifier)

        # Verify the snapshot size same as the volume
        identifier = {'name': 'snapshot-%s' % snapshot['id']}
        snap_size = (self.driver.manage_existing_snapshot_get_size(
                     snapshot, identifier))
        self.assertEqual(snap_size, volume['size'])
        self.driver.delete_volume(volume)



# Copyright (c) 2014 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Classes and utility methods for datastore selection.
"""

import random

from oslo_log import log as logging
from oslo_vmware import pbm
from oslo_vmware import vim_util

from cinder import coordination
from cinder.volume.drivers.vmware import exceptions as vmdk_exceptions


LOG = logging.getLogger(__name__)


class DatastoreType(object):
    """Supported datastore types."""

    NFS = "nfs"
    VMFS = "vmfs"
    VSAN = "vsan"
    VVOL = "vvol"
    NFS41 = "nfs41"

    _ALL_TYPES = {NFS, VMFS, VSAN, VVOL, NFS41}

    @staticmethod
    def get_all_types():
        return DatastoreType._ALL_TYPES


class DatastoreSelector(object):
    """Class for selecting datastores which satisfy input requirements."""

    HARD_AFFINITY_DS_TYPE = "hardAffinityDatastoreTypes"
    HARD_ANTI_AFFINITY_DS = "hardAntiAffinityDatastores"
    SIZE_BYTES = "sizeBytes"
    PROFILE_NAME = "storageProfileName"

    # TODO(vbala) Remove dependency on volumeops.
    def __init__(self, vops, session, max_objects, ds_regex=None):
        self._vops = vops
        self._session = session
        self._max_objects = max_objects
        self._ds_regex = ds_regex
        self._profile_id_cache = {}

    @coordination.synchronized('vmware-datastore-profile-{profile_name}')
    def get_profile_id(self, profile_name):
        """Get vCenter profile ID for the given profile name.

        :param profile_name: profile name
        :return: vCenter profile ID
        :raises ProfileNotFoundException:
        """
        if profile_name in self._profile_id_cache:
            LOG.debug("Returning cached ID for profile: %s.", profile_name)
            return self._profile_id_cache[profile_name]

        profile_id = pbm.get_profile_id_by_name(self._session, profile_name)
        if profile_id is None:
            LOG.error("Storage profile: %s cannot be found in vCenter.",
                      profile_name)
            raise vmdk_exceptions.ProfileNotFoundException(
                storage_profile=profile_name)

        self._profile_id_cache[profile_name] = profile_id
        LOG.debug("Storage profile: %(name)s resolved to vCenter profile ID: "
                  "%(id)s.",
                  {'name': profile_name,
                   'id': profile_id})
        return profile_id

    def _filter_by_profile(self, datastores, profile_id):
        """Filter out input datastores that do not match the given profile."""
        cf = self._session.pbm.client.factory
        hubs = pbm.convert_datastores_to_hubs(cf, datastores)
        hubs = pbm.filter_hubs_by_profile(self._session, hubs, profile_id)
        hub_ids = [hub.hubId for hub in hubs]
        return {k: v for k, v in datastores.items()
                if vim_util.get_moref_value(k) in hub_ids}

    def _filter_datastores(self,
                           datastores,
                           size_bytes,
                           profile_id,
                           hard_anti_affinity_ds,
                           hard_affinity_ds_types,
                           valid_host_refs=None):

        if not datastores:
            return

        def _is_valid_ds_type(summary):
            ds_type = summary.type.lower()
            return (ds_type in DatastoreType.get_all_types() and
                    (hard_affinity_ds_types is None or
                     ds_type in hard_affinity_ds_types))

        def _is_ds_usable(summary):
            return summary.accessible and not self._vops._in_maintenance(
                summary)

        valid_host_refs = valid_host_refs or []
        valid_hosts = [vim_util.get_moref_value(host_ref)
                       for host_ref in valid_host_refs]

        def _is_ds_accessible_to_valid_host(host_mounts):
            for host_mount in host_mounts:
                if vim_util.get_moref_value(host_mount.key) in valid_hosts:
                    return True

        def _is_ds_valid(ds_ref, ds_props):
            summary = ds_props.get('summary')
            host_mounts = ds_props.get('host')
            if (summary is None or host_mounts is None):
                return False

            if self._ds_regex and not self._ds_regex.match(summary.name):
                return False

            if (hard_anti_affinity_ds and
                    vim_util.get_moref_value(ds_ref) in hard_anti_affinity_ds):
                return False

            if summary.capacity == 0 or summary.freeSpace < size_bytes:
                return False

            if (valid_hosts and
                    not _is_ds_accessible_to_valid_host(host_mounts)):
                return False

            return _is_valid_ds_type(summary) and _is_ds_usable(summary)

        datastores = {k: v for k, v in datastores.items()
                      if _is_ds_valid(k, v)}

        if datastores and profile_id:
            datastores = self._filter_by_profile(datastores, profile_id)

        return datastores

    def _get_object_properties(self, obj_content):
        props = {}
        if hasattr(obj_content, 'propSet'):
            prop_set = obj_content.propSet
            if prop_set:
                props = {prop.name: prop.val for prop in prop_set}
        return props

    def _get_datastores(self):
        datastores = {}
        retrieve_result = self._session.invoke_api(
            vim_util,
            'get_objects',
            self._session.vim,
            'Datastore',
            self._max_objects,
            properties_to_collect=['host', 'summary'])

        while retrieve_result:
            if retrieve_result.objects:
                for obj_content in retrieve_result.objects:
                    props = self._get_object_properties(obj_content)
                    if ('host' in props and
                            hasattr(props['host'], 'DatastoreHostMount')):
                        props['host'] = props['host'].DatastoreHostMount
                    datastores[obj_content.obj] = props
            retrieve_result = self._session.invoke_api(vim_util,
                                                       'continue_retrieval',
                                                       self._session.vim,
                                                       retrieve_result)

        return datastores

    def _get_host_properties(self, host_ref):
        retrieve_result = self._session.invoke_api(vim_util,
                                                   'get_object_properties',
                                                   self._session.vim,
                                                   host_ref,
                                                   ['runtime', 'parent'])

        if retrieve_result:
            return self._get_object_properties(retrieve_result[0])

    def _get_resource_pool(self, cluster_ref):
        return self._session.invoke_api(vim_util,
                                        'get_object_property',
                                        self._session.vim,
                                        cluster_ref,
                                        'resourcePool')

    def _select_best_datastore(self, datastores, valid_host_refs=None):

        if not datastores:
            return

        def _sort_key(ds_props):
            host = ds_props.get('host')
            summary = ds_props.get('summary')
            space_utilization = (1.0 -
                                 (summary.freeSpace / float(summary.capacity)))
            return (-len(host), space_utilization)

        host_prop_map = {}

        def _is_host_usable(host_ref):
            props = host_prop_map.get(vim_util.get_moref_value(host_ref))
            if props is None:
                props = self._get_host_properties(host_ref)
                host_prop_map[vim_util.get_moref_value(host_ref)] = props

            runtime = props.get('runtime')
            parent = props.get('parent')
            if runtime and parent:
                return (runtime.connectionState == 'connected' and
                        not runtime.inMaintenanceMode)
            else:
                return False

        valid_host_refs = valid_host_refs or []
        valid_hosts = [vim_util.get_moref_value(host_ref)
                       for host_ref in valid_host_refs]

        def _select_host(host_mounts):
            random.shuffle(host_mounts)
            for host_mount in host_mounts:
                host_mount_key_value = vim_util.get_moref_value(host_mount.key)
                if valid_hosts and host_mount_key_value not in valid_hosts:
                    continue
                if (self._vops._is_usable(host_mount.mountInfo) and
                        _is_host_usable(host_mount.key)):
                    return host_mount.key

        sorted_ds_props = sorted(datastores.values(), key=_sort_key)
        for ds_props in sorted_ds_props:
            host_ref = _select_host(ds_props['host'])
            if host_ref:
                host_ref_value = vim_util.get_moref_value(host_ref)
                rp = self._get_resource_pool(
                    host_prop_map[host_ref_value]['parent'])
                return (host_ref, rp, ds_props['summary'])

    def select_datastore(self, req, hosts=None):
        """Selects a datastore satisfying the given requirements.

        A datastore which is connected to maximum number of hosts is
        selected. Ties if any are broken based on space utilization--
        datastore with least space utilization is preferred. It returns
        the selected datastore's summary along with a host and resource
        pool where the volume can be created.

        :param req: selection requirements
        :param hosts: list of hosts to consider
        :return: (host, resourcePool, summary)
        """
        LOG.debug("Using requirements: %s for datastore selection.", req)

        hard_affinity_ds_types = req.get(
            DatastoreSelector.HARD_AFFINITY_DS_TYPE)
        hard_anti_affinity_datastores = req.get(
            DatastoreSelector.HARD_ANTI_AFFINITY_DS)
        size_bytes = req[DatastoreSelector.SIZE_BYTES]
        profile_name = req.get(DatastoreSelector.PROFILE_NAME)

        profile_id = None
        if profile_name is not None:
            profile_id = self.get_profile_id(profile_name)

        datastores = self._get_datastores()
        datastores = self._filter_datastores(datastores,
                                             size_bytes,
                                             profile_id,
                                             hard_anti_affinity_datastores,
                                             hard_affinity_ds_types,
                                             valid_host_refs=hosts)
        res = self._select_best_datastore(datastores, valid_host_refs=hosts)
        LOG.debug("Selected (host, resourcepool, datastore): %s", res)
        return res

    def is_datastore_compliant(self, datastore, profile_name):
        """Check if the datastore is compliant with given profile.

        :param datastore: datastore to check the compliance
        :param profile_name: profile to check the compliance against
        :return: True if the datastore is compliant; False otherwise
        :raises ProfileNotFoundException:
        """
        LOG.debug("Checking datastore: %(datastore)s compliance against "
                  "profile: %(profile)s.",
                  {'datastore': datastore,
                   'profile': profile_name})
        if profile_name is None:
            # Any datastore is trivially compliant with a None profile.
            return True

        profile_id = self.get_profile_id(profile_name)
        # _filter_by_profile expects a map of datastore references to its
        # properties. It only uses the properties to construct a map of
        # filtered datastores to its properties. Here we don't care about
        # the datastore property, so pass it as None.
        is_compliant = bool(self._filter_by_profile({datastore: None},
                                                    profile_id))
        LOG.debug("Compliance is %(is_compliant)s for datastore: "
                  "%(datastore)s against profile: %(profile)s.",
                  {'is_compliant': is_compliant,
                   'datastore': datastore,
                   'profile': profile_name})
        return is_compliant



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import unittest
from unittest import mock

from cinder.objects import fields
from cinder.tests.unit.volume.drivers.dell_emc.unity \
    import fake_exception as ex
from cinder.tests.unit.volume.drivers.dell_emc.unity import test_adapter
from cinder.volume import configuration as conf
from cinder.volume.drivers.dell_emc.unity import driver


########################
#
#   Start of Mocks
#
########################

class MockAdapter(object):
    def __init__(self):
        self.is_setup = False

    def do_setup(self, driver_object, configuration):
        self.is_setup = True
        raise ex.AdapterSetupError()

    @staticmethod
    def create_volume(volume):
        return volume

    @staticmethod
    def create_volume_from_snapshot(volume, snapshot):
        return volume

    @staticmethod
    def create_cloned_volume(volume, src_vref):
        return volume

    @staticmethod
    def extend_volume(volume, new_size):
        volume.size = new_size

    @staticmethod
    def delete_volume(volume):
        volume.exists = False

    @staticmethod
    def create_snapshot(snapshot):
        snapshot.exists = True
        return snapshot

    @staticmethod
    def delete_snapshot(snapshot):
        snapshot.exists = False

    @staticmethod
    def initialize_connection(volume, connector):
        return {'volume': volume, 'connector': connector}

    @staticmethod
    def terminate_connection(volume, connector):
        return {'volume': volume, 'connector': connector}

    @staticmethod
    def update_volume_stats():
        return {'stats': 123}

    @staticmethod
    def manage_existing(volume, existing_ref):
        volume.managed = True
        return volume

    @staticmethod
    def manage_existing_get_size(volume, existing_ref):
        volume.managed = True
        volume.size = 7
        return volume

    @staticmethod
    def get_pool_name(volume):
        return 'pool_0'

    @staticmethod
    def initialize_connection_snapshot(snapshot, connector):
        return {'snapshot': snapshot, 'connector': connector}

    @staticmethod
    def terminate_connection_snapshot(snapshot, connector):
        return {'snapshot': snapshot, 'connector': connector}

    @staticmethod
    def restore_snapshot(volume, snapshot):
        return True

    @staticmethod
    def migrate_volume(volume, host):
        return True, {}

    @staticmethod
    def create_group(group):
        return group

    @staticmethod
    def delete_group(group):
        return group

    @staticmethod
    def update_group(group, add_volumes, remove_volumes):
        return group, add_volumes, remove_volumes

    @staticmethod
    def create_group_from_snap(group, volumes, group_snapshot, snapshots):
        return group, volumes, group_snapshot, snapshots

    @staticmethod
    def create_cloned_group(group, volumes, source_group, source_vols):
        return group, volumes, source_group, source_vols

    @staticmethod
    def create_group_snapshot(group_snapshot, snapshots):
        return group_snapshot, snapshots

    @staticmethod
    def delete_group_snapshot(group_snapshot):
        return group_snapshot

    def failover(self, volumes, secondary_id=None, groups=None):
        return {'volumes': volumes,
                'secondary_id': secondary_id,
                'groups': groups}

    @staticmethod
    def enable_replication(context, group, volumes):
        if volumes and group:
            return {'replication_status':
                    fields.ReplicationStatus.ENABLED}, None
        return {}, None

    @staticmethod
    def disable_replication(context, group, volumes):
        if volumes and group:
            return {'replication_status':
                    fields.ReplicationStatus.DISABLED}, None
        return {}, None

    @staticmethod
    def failover_replication(context, group, volumes,
                             secondary_backend_id):
        group_update = {}
        volumes_update = []
        if volumes and group and secondary_backend_id:
            group_update = {'replication_status':
                            fields.ReplicationStatus.FAILED_OVER}
            for volume in volumes:
                volume_update = {
                    'id': volume.id,
                    'replication_status':
                        fields.ReplicationStatus.FAILED_OVER}
                volumes_update.append(volume_update)
            return group_update, volumes_update
        return group_update, None

    def retype(self, ctxt, volume, new_type, diff, host):
        return True


class MockReplicationManager(object):
    def __init__(self):
        self.active_adapter = MockAdapter()

    def do_setup(self, d):
        if isinstance(d, driver.UnityDriver):
            raise ex.ReplicationManagerSetupError()


########################
#
#   Start of Tests
#
########################


patch_check_cg = mock.patch(
    'cinder.volume.volume_utils.is_group_a_cg_snapshot_type',
    side_effect=lambda g: not g.id.endswith('_generic'))


class UnityDriverTest(unittest.TestCase):
    @staticmethod
    def get_volume():
        return test_adapter.MockOSResource(provider_location='id^lun_43',
                                           id='id_43')

    @staticmethod
    def get_volumes():
        volumes = []
        for number in ['50', '51', '52', '53']:
            volume = test_adapter.MockOSResource(
                provider_location='id^lun_' + number, id='id_' + number)
            volumes.append(volume)
        return volumes

    @staticmethod
    def get_generic_group():
        return test_adapter.MockOSResource(name='group_name_generic',
                                           id='group_id_generic')

    @staticmethod
    def get_cg():
        return test_adapter.MockOSResource(name='group_name_cg',
                                           id='group_id_cg')

    @classmethod
    def get_snapshot(cls):
        return test_adapter.MockOSResource(volume=cls.get_volume())

    @classmethod
    def get_generic_group_snapshot(cls):
        return test_adapter.MockOSResource(group=cls.get_generic_group(),
                                           id='group_snapshot_id_generic')

    @classmethod
    def get_cg_group_snapshot(cls):
        return test_adapter.MockOSResource(group=cls.get_cg(),
                                           id='group_snapshot_id_cg')

    @staticmethod
    def get_context():
        return None

    @staticmethod
    def get_connector():
        return {'host': 'host1'}

    def setUp(self):
        self.config = conf.Configuration(None)
        self.driver = driver.UnityDriver(configuration=self.config)
        self.driver.replication_manager = MockReplicationManager()

    def test_default_initialize(self):
        config = conf.Configuration(None)
        iscsi_driver = driver.UnityDriver(configuration=config)
        self.assertListEqual([], config.unity_storage_pool_names)
        self.assertListEqual([], config.unity_io_ports)
        self.assertTrue(config.san_thin_provision)
        self.assertEqual('', config.san_ip)
        self.assertEqual('admin', config.san_login)
        self.assertEqual('', config.san_password)
        self.assertEqual('', config.san_private_key)
        self.assertEqual('', config.san_clustername)
        self.assertEqual(22, config.san_ssh_port)
        self.assertEqual(False, config.san_is_local)
        self.assertEqual(30, config.ssh_conn_timeout)
        self.assertEqual(1, config.ssh_min_pool_conn)
        self.assertEqual(5, config.ssh_max_pool_conn)
        self.assertEqual('iSCSI', iscsi_driver.protocol)
        self.assertIsNone(iscsi_driver.active_backend_id)

    def test_initialize_with_active_backend_id(self):
        config = conf.Configuration(None)
        iscsi_driver = driver.UnityDriver(configuration=config,
                                          active_backend_id='secondary_unity')
        self.assertEqual('secondary_unity', iscsi_driver.active_backend_id)

    def test_fc_initialize(self):
        config = conf.Configuration(None)
        config.storage_protocol = 'fc'
        fc_driver = driver.UnityDriver(configuration=config)
        self.assertEqual('FC', fc_driver.protocol)

    def test_do_setup(self):
        def f():
            self.driver.do_setup(None)

        self.assertRaises(ex.ReplicationManagerSetupError, f)

    def test_create_volume(self):
        volume = self.get_volume()
        self.assertEqual(volume, self.driver.create_volume(volume))

    def test_create_volume_from_snapshot(self):
        volume = self.get_volume()
        snap = self.get_snapshot()
        self.assertEqual(
            volume, self.driver.create_volume_from_snapshot(volume, snap))

    def test_create_cloned_volume(self):
        volume = self.get_volume()
        self.assertEqual(
            volume, self.driver.create_cloned_volume(volume, None))

    def test_extend_volume(self):
        volume = self.get_volume()
        self.driver.extend_volume(volume, 6)
        self.assertEqual(6, volume.size)

    def test_delete_volume(self):
        volume = self.get_volume()
        self.driver.delete_volume(volume)
        self.assertFalse(volume.exists)

    def test_migrate_volume(self):
        volume = self.get_volume()
        ret = self.driver.migrate_volume(self.get_context(),
                                         volume,
                                         'HostA@BackendB#PoolC')
        self.assertEqual((True, {}), ret)

    def test_retype_volume(self):
        volume = self.get_volume()
        new_type = {'name': u'type01', 'qos_specs_id': 'test_qos_id',
                    'extra_specs': {},
                    'id': u'd67c4480-a61b-44c0-a58b-24c0357cadeb'}
        diff = None
        ret = self.driver.retype(self.get_context(),
                                 volume, new_type, diff,
                                 'HostA@BackendB#PoolC')
        self.assertTrue(ret)

    def test_create_snapshot(self):
        snapshot = self.get_snapshot()
        self.driver.create_snapshot(snapshot)
        self.assertTrue(snapshot.exists)

    def test_delete_snapshot(self):
        snapshot = self.get_snapshot()
        self.driver.delete_snapshot(snapshot)
        self.assertFalse(snapshot.exists)

    def test_ensure_export(self):
        self.assertIsNone(self.driver.ensure_export(
            self.get_context(), self.get_volume()))

    def test_create_export(self):
        self.assertIsNone(self.driver.create_export(
            self.get_context(), self.get_volume(), self.get_connector()))

    def test_remove_export(self):
        self.assertIsNone(self.driver.remove_export(
            self.get_context(), self.get_volume()))

    def test_check_for_export(self):
        self.assertIsNone(self.driver.check_for_export(
            self.get_context(), self.get_volume()))

    def test_initialize_connection(self):
        volume = self.get_volume()
        connector = self.get_connector()
        conn_info = self.driver.initialize_connection(volume, connector)
        self.assertEqual(volume, conn_info['volume'])
        self.assertEqual(connector, conn_info['connector'])

    def test_terminate_connection(self):
        volume = self.get_volume()
        connector = self.get_connector()
        conn_info = self.driver.terminate_connection(volume, connector)
        self.assertEqual(volume, conn_info['volume'])
        self.assertEqual(connector, conn_info['connector'])

    def test_update_volume_stats(self):
        stats = self.driver.get_volume_stats(True)
        self.assertEqual(123, stats['stats'])
        self.assertEqual(self.driver.VERSION, stats['driver_version'])
        self.assertEqual(self.driver.VENDOR, stats['vendor_name'])

    def test_manage_existing(self):
        volume = self.driver.manage_existing(self.get_volume(), None)
        self.assertTrue(volume.managed)

    def test_manage_existing_get_size(self):
        volume = self.driver.manage_existing_get_size(self.get_volume(), None)
        self.assertTrue(volume.managed)
        self.assertEqual(7, volume.size)

    def test_get_pool(self):
        self.assertEqual('pool_0', self.driver.get_pool(self.get_volume()))

    def test_unmanage(self):
        ret = self.driver.unmanage(None)
        self.assertIsNone(ret)

    def test_backup_use_temp_snapshot(self):
        self.assertTrue(self.driver.backup_use_temp_snapshot())

    def test_initialize_connection_snapshot(self):
        snapshot = self.get_snapshot()
        conn_info = self.driver.initialize_connection_snapshot(
            snapshot, self.get_connector())
        self.assertEqual(snapshot, conn_info['snapshot'])

    def test_terminate_connection_snapshot(self):
        snapshot = self.get_snapshot()
        conn_info = self.driver.terminate_connection_snapshot(
            snapshot, self.get_connector())
        self.assertEqual(snapshot, conn_info['snapshot'])

    def test_restore_snapshot(self):
        snapshot = self.get_snapshot()
        volume = self.get_volume()
        r = self.driver.revert_to_snapshot(None, volume, snapshot)
        self.assertTrue(r)

    @patch_check_cg
    def test_operate_generic_group_not_implemented(self, _):
        group = self.get_generic_group()
        context = self.get_context()

        for func in (self.driver.create_group, self.driver.update_group):
            self.assertRaises(NotImplementedError,
                              functools.partial(func, context, group))

        volumes = [self.get_volume()]
        for func in (self.driver.delete_group,
                     self.driver.create_group_from_src):
            self.assertRaises(NotImplementedError,
                              functools.partial(func, context, group, volumes))

        group_snap = self.get_generic_group_snapshot()
        volume_snaps = [self.get_snapshot()]
        for func in (self.driver.create_group_snapshot,
                     self.driver.delete_group_snapshot):
            self.assertRaises(NotImplementedError,
                              functools.partial(func, context, group_snap,
                                                volume_snaps))

    @patch_check_cg
    def test_create_group_cg(self, _):
        cg = self.get_cg()
        ret = self.driver.create_group(self.get_context(), cg)
        self.assertEqual(ret, cg)

    @patch_check_cg
    def test_delete_group_cg(self, _):
        cg = self.get_cg()
        volumes = [self.get_volume()]
        ret = self.driver.delete_group(self.get_context(), cg, volumes)
        self.assertEqual(ret, cg)

    @patch_check_cg
    def test_update_group_cg(self, _):
        cg = self.get_cg()
        volumes = [self.get_volume()]
        ret = self.driver.update_group(self.get_context(), cg,
                                       add_volumes=volumes)
        self.assertEqual(ret[0], cg)
        self.assertListEqual(ret[1], volumes)
        self.assertIsNone(ret[2])

    @patch_check_cg
    def test_create_group_from_src_group(self, _):
        cg = self.get_cg()
        volumes = [self.get_volume()]
        source_group = cg
        ret = self.driver.create_group_from_src(self.get_context(), cg,
                                                volumes,
                                                source_group=source_group)
        self.assertEqual(ret[0], cg)
        self.assertListEqual(ret[1], volumes)
        self.assertEqual(ret[2], source_group)
        self.assertIsNone(ret[3])

    @patch_check_cg
    def test_create_group_from_src_group_snapshot(self, _):
        cg = self.get_cg()
        volumes = [self.get_volume()]
        cg_snap = self.get_cg_group_snapshot()
        ret = self.driver.create_group_from_src(self.get_context(), cg,
                                                volumes,
                                                group_snapshot=cg_snap)
        self.assertEqual(ret[0], cg)
        self.assertListEqual(ret[1], volumes)
        self.assertEqual(ret[2], cg_snap)
        self.assertIsNone(ret[3])

    @patch_check_cg
    def test_create_group_snapshot_cg(self, _):
        cg_snap = self.get_cg_group_snapshot()
        ret = self.driver.create_group_snapshot(self.get_context(), cg_snap,
                                                None)
        self.assertEqual(ret[0], cg_snap)
        self.assertIsNone(ret[1])

    @patch_check_cg
    def test_delete_group_snapshot_cg(self, _):
        cg_snap = self.get_cg_group_snapshot()
        ret = self.driver.delete_group_snapshot(self.get_context(), cg_snap,
                                                None)
        self.assertEqual(ret, cg_snap)

    def test_failover_host(self):
        volume = self.get_volume()
        called = self.driver.failover_host(None, [volume],
                                           secondary_id='secondary_unity',
                                           groups=None)
        self.assertListEqual(called['volumes'], [volume])
        self.assertEqual('secondary_unity', called['secondary_id'])
        self.assertIsNone(called['groups'])

    def test_enable_replication(self):
        cg = self.get_cg()
        volumes = self.get_volumes()
        result = self.driver.enable_replication(None, cg, volumes)
        self.assertEqual(result,
                         ({'replication_status':
                          fields.ReplicationStatus.ENABLED},
                          None))

    def test_disable_replication(self):
        cg = self.get_cg()
        volumes = self.get_volumes()
        result = self.driver.disable_replication(None, cg, volumes)
        self.assertEqual(result,
                         ({'replication_status':
                          fields.ReplicationStatus.DISABLED},
                          None))

    def test_failover_replication(self):
        cg = self.get_cg()
        volumes = self.get_volumes()
        result = self.driver.failover_replication(
            None, cg, volumes, 'test_secondary_id')
        volumes = [{'id': 'id_50', 'replication_status': 'failed-over'},
                   {'id': 'id_51', 'replication_status': 'failed-over'},
                   {'id': 'id_52', 'replication_status': 'failed-over'},
                   {'id': 'id_53', 'replication_status': 'failed-over'}]

        self.assertEqual(result,
                         ({'replication_status':
                          fields.ReplicationStatus.FAILED_OVER},
                          volumes))



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.objects import fields as c_fields


@base.CinderObjectRegistry.register
class VolumeAttachment(base.CinderPersistentObject, base.CinderObject,
                       base.CinderObjectDictCompat,
                       base.CinderComparableObject):
    # Version 1.0: Initial version
    # Version 1.1: Added volume relationship
    # Version 1.2: Added connection_info attribute
    # Version 1.3: Added the connector attribute.
    VERSION = '1.3'

    OPTIONAL_FIELDS = ['volume']
    obj_extra_fields = ['project_id', 'volume_host']

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        'id': fields.UUIDField(),
        'volume_id': fields.UUIDField(),
        'instance_uuid': fields.UUIDField(nullable=True),
        'attached_host': fields.StringField(nullable=True),
        'mountpoint': fields.StringField(nullable=True),

        'attach_time': fields.DateTimeField(nullable=True),
        'detach_time': fields.DateTimeField(nullable=True),

        'attach_status': c_fields.VolumeAttachStatusField(nullable=True),
        'attach_mode': fields.StringField(nullable=True),

        'volume': fields.ObjectField('Volume', nullable=False),
        'connection_info': c_fields.DictOfNullableField(nullable=True),
        'connector': c_fields.DictOfNullableField(nullable=True)
    }

    @property
    def project_id(self):
        return self.volume.project_id

    @property
    def volume_host(self):
        return self.volume.host

    @classmethod
    def _get_expected_attrs(cls, context, *args, **kwargs):
        return ['volume']

    @classmethod
    def _from_db_object(cls, context, attachment, db_attachment,
                        expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = cls._get_expected_attrs(context)

        for name, field in attachment.fields.items():
            if name in cls.OPTIONAL_FIELDS:
                continue
            value = db_attachment.get(name)
            if isinstance(field, fields.IntegerField):
                value = value or 0
            if name in ('connection_info', 'connector'):
                # Both of these fields are nullable serialized json dicts.
                setattr(attachment, name,
                        jsonutils.loads(value) if value else None)
            else:
                attachment[name] = value
        # NOTE: Check against the ORM instance's dictionary instead of using
        # hasattr or get to avoid the lazy loading of the Volume on
        # VolumeList.get_all.
        # Getting a Volume loads its VolumeAttachmentList, which think they
        # have the volume loaded, but they don't.  More detail on
        # https://review.opendev.org/632549
        # and its related bug report.
        if 'volume' in expected_attrs and 'volume' in vars(db_attachment):
            db_volume = db_attachment.volume
            if db_volume:
                attachment.volume = objects.Volume._from_db_object(
                    context, objects.Volume(), db_volume)

        attachment._context = context
        attachment.obj_reset_changes()

        # This is an online data migration which we should remove when enough
        # time has passed and we have a blocker schema migration to check to
        # make sure that the attachment_specs table is empty. Operators should
        # run the "cinder-manage db online_data_migrations" CLI to force the
        # migration on-demand.
        connector = db.attachment_specs_get(context, attachment.id)
        if connector:
            # Update ourselves and delete the attachment_specs.
            attachment.connector = connector
            attachment.save()
            # TODO(mriedem): Really need a delete-all method for this.
            for spec_key in connector:
                db.attachment_specs_delete(
                    context, attachment.id, spec_key)

        return attachment

    def obj_load_attr(self, attrname):
        if attrname not in self.OPTIONAL_FIELDS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        if attrname == 'volume':
            volume = objects.Volume.get_by_id(self._context, self.volume_id)
            self.volume = volume

        self.obj_reset_changes(fields=[attrname])

    @staticmethod
    def _convert_connection_info_to_db_format(updates):
        properties = updates.pop('connection_info', None)
        if properties is not None:
            updates['connection_info'] = jsonutils.dumps(properties)

    @staticmethod
    def _convert_connector_to_db_format(updates):
        connector = updates.pop('connector', None)
        if connector is not None:
            updates['connector'] = jsonutils.dumps(connector)

    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            if 'connection_info' in updates:
                self._convert_connection_info_to_db_format(updates)
            if 'connector' in updates:
                self._convert_connector_to_db_format(updates)
            if 'volume' in updates:
                raise exception.ObjectActionError(action='save',
                                                  reason=_('volume changed'))

            db.volume_attachment_update(self._context, self.id, updates)
            self.obj_reset_changes()

    def finish_attach(self, instance_uuid, host_name,
                      mount_point, attach_mode='rw'):
        with self.obj_as_admin():
            db_volume, updated_values = db.volume_attached(
                self._context, self.id,
                instance_uuid, host_name,
                mount_point, attach_mode)
        self.update(updated_values)
        self.obj_reset_changes(updated_values.keys())
        return objects.Volume._from_db_object(self._context,
                                              objects.Volume(),
                                              db_volume)

    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.cinder_obj_get_changes()
        if 'connector' in updates:
            self._convert_connector_to_db_format(updates)
        with self.obj_as_admin():
            db_attachment = db.volume_attach(self._context, updates)
        self._from_db_object(self._context, self, db_attachment)

    def destroy(self):
        updated_values = db.attachment_destroy(self._context, self.id)
        self.update(updated_values)
        self.obj_reset_changes(updated_values.keys())


@base.CinderObjectRegistry.register
class VolumeAttachmentList(base.ObjectListBase, base.CinderObject):
    # Version 1.0: Initial version
    # Version 1.1: Remove volume_id in get_by_host|instance
    VERSION = '1.1'

    fields = {
        'objects': fields.ListOfObjectsField('VolumeAttachment'),
    }

    @classmethod
    def get_all_by_volume_id(cls, context, volume_id):
        attachments = db.volume_attachment_get_all_by_volume_id(context,
                                                                volume_id)
        return base.obj_make_list(context,
                                  cls(context),
                                  objects.VolumeAttachment,
                                  attachments)

    @classmethod
    def get_all_by_host(cls, context, host, search_opts=None):
        attachments = db.volume_attachment_get_all_by_host(context,
                                                           host,
                                                           search_opts)
        return base.obj_make_list(context, cls(context),
                                  objects.VolumeAttachment, attachments)

    @classmethod
    def get_all_by_instance_uuid(cls, context,
                                 instance_uuid, search_opts=None):
        attachments = db.volume_attachment_get_all_by_instance_uuid(
            context, instance_uuid, search_opts)
        return base.obj_make_list(context, cls(context),
                                  objects.VolumeAttachment, attachments)

    @classmethod
    def get_all(cls, context, search_opts=None,
                marker=None, limit=None, offset=None,
                sort_keys=None, sort_direction=None):
        attachments = db.volume_attachment_get_all(
            context, search_opts, marker, limit, offset, sort_keys,
            sort_direction)
        return base.obj_make_list(context, cls(context),
                                  objects.VolumeAttachment, attachments)

    @classmethod
    def get_all_by_project(cls, context, project_id, search_opts=None,
                           marker=None, limit=None, offset=None,
                           sort_keys=None, sort_direction=None):
        attachments = db.volume_attachment_get_all_by_project(
            context, project_id, search_opts, marker, limit, offset, sort_keys,
            sort_direction)
        return base.obj_make_list(context, cls(context),
                                  objects.VolumeAttachment, attachments)



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

DEFAULT_TIMEOUT = 50
LOGIN_SOCKET_TIMEOUT = 32

CONNECT_ERROR = 403
ERROR_UNAUTHORIZED = 10000003
VOLUME_NOT_EXIST = (31000000, 50150005)

BASIC_URI = '/dsware/service/'
CONF_PATH = "/etc/cinder/cinder.conf"

CONF_ADDRESS = "dsware_rest_url"
CONF_MANAGER_IP = "manager_ips"
CONF_POOLS = "dsware_storage_pools"
CONF_PWD = "san_password"
CONF_USER = "san_login"



#
# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for cinder.api.contrib.quotas.py"""

from unittest import mock
import uuid

from oslo_config import cfg
from oslo_config import fixture as config_fixture
import webob.exc

from cinder.api.contrib import quotas
from cinder import context
from cinder import db
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import test_db_api


CONF = cfg.CONF


def make_body(root=True, gigabytes=1000, snapshots=10,
              volumes=10, backups=10, backup_gigabytes=1000,
              tenant_id=fake.PROJECT_ID, per_volume_gigabytes=-1, groups=10,
              subproject=False):
    resources = {'gigabytes': gigabytes,
                 'snapshots': snapshots,
                 'volumes': volumes,
                 'backups': backups,
                 'backup_gigabytes': backup_gigabytes,
                 'per_volume_gigabytes': per_volume_gigabytes,
                 'groups': groups}
    # need to consider preexisting volume types as well
    volume_types = db.volume_type_get_all(context.get_admin_context())

    for volume_type in volume_types:
        # default values for subproject are 0
        quota = 0 if subproject else -1
        resources['gigabytes_' + volume_type] = quota
        resources['snapshots_' + volume_type] = quota
        resources['volumes_' + volume_type] = quota

    if tenant_id:
        resources['id'] = tenant_id
    if root:
        result = {'quota_set': resources}
    else:
        result = resources
    return result


def make_subproject_body(root=True, gigabytes=0, snapshots=0,
                         volumes=0, backups=0, backup_gigabytes=0,
                         tenant_id=fake.PROJECT_ID, per_volume_gigabytes=0):
    return make_body(root=root, gigabytes=gigabytes, snapshots=snapshots,
                     volumes=volumes, backups=backups,
                     backup_gigabytes=backup_gigabytes, tenant_id=tenant_id,
                     per_volume_gigabytes=per_volume_gigabytes,
                     subproject=True)


class QuotaSetsControllerTestBase(test.TestCase):

    class FakeProject(object):

        def __init__(self, id=fake.PROJECT_ID, parent_id=None,
                     is_admin_project=False):
            self.id = id
            self.parent_id = parent_id
            self.subtree = None
            self.parents = None
            self.is_admin_project = is_admin_project

    def setUp(self):
        super(QuotaSetsControllerTestBase, self).setUp()

        self.controller = quotas.QuotaSetsController()

        self.req = mock.Mock()
        self.req.environ = {'cinder.context': context.get_admin_context()}
        self.req.environ['cinder.context'].is_admin = True
        self.req.params = {}

        self.req.environ['cinder.context'].project_id = uuid.uuid4().hex

        get_patcher = mock.patch('cinder.api.api_utils.get_project',
                                 self._get_project)
        get_patcher.start()
        self.addCleanup(get_patcher.stop)

        self.auth_url = 'http://localhost:5000'
        self.fixture = self.useFixture(config_fixture.Config(CONF))
        self.fixture.config(auth_url=self.auth_url, group='keystone_authtoken')

    def _get_project(self, context, id, subtree_as_ids=False,
                     parents_as_ids=False, is_admin_project=False):
        return self.project_by_id.get(id, self.FakeProject())

    def _create_fake_quota_usages(self, usage_map):
        self._fake_quota_usages = {}
        for key, val in usage_map.items():
            self._fake_quota_usages[key] = {'in_use': val}

    def _fake_quota_usage_get_all_by_project(self, context, project_id):
        return {'volumes': self._fake_quota_usages[project_id]}


class QuotaSetsControllerTest(QuotaSetsControllerTestBase):
    def test_defaults(self):
        result = self.controller.defaults(self.req, fake.PROJECT_ID)
        self.assertDictEqual(make_body(), result)

    def test_show(self):
        result = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(make_body(), result)

    def test_show_not_authorized(self):
        self.req.environ['cinder.context'].is_admin = False
        self.req.environ['cinder.context'].user_id = fake.USER_ID
        self.req.environ['cinder.context'].project_id = fake.PROJECT_ID
        self.req.environ['cinder.context'].roles = ['member', 'reader']
        self.assertRaises(exception.PolicyNotAuthorized, self.controller.show,
                          self.req, fake.PROJECT2_ID)

    def test_show_non_admin_user(self):
        self.controller._get_quotas = mock.Mock(side_effect=
                                                self.controller._get_quotas)
        result = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(make_body(), result)
        self.controller._get_quotas.assert_called_with(
            self.req.environ['cinder.context'], fake.PROJECT_ID, False)

    def test_show_with_invalid_usage_param(self):
        self.req.params = {'usage': 'InvalidBool'}
        self.assertRaises(exception.InvalidParameterValue,
                          self.controller.show,
                          self.req, fake.PROJECT2_ID)

    def test_show_with_valid_usage_param(self):
        self.req.params = {'usage': 'false'}
        result = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(make_body(), result)

    def test_update(self):
        body = make_body(gigabytes=2000, snapshots=15,
                         volumes=5, backups=5, tenant_id=None)
        result = self.controller.update(self.req, fake.PROJECT_ID, body=body)
        self.assertDictEqual(body, result)

        body = make_body(gigabytes=db.MAX_INT, tenant_id=None)
        result = self.controller.update(self.req, fake.PROJECT_ID, body=body)
        self.assertDictEqual(body, result)

    @mock.patch(
        'cinder.api.openstack.wsgi.Controller.validate_string_length')
    def test_update_limit(self, mock_validate):
        body = {'quota_set': {'volumes': 10}}
        result = self.controller.update(self.req, fake.PROJECT_ID, body=body)

        self.assertEqual(10, result['quota_set']['volumes'])
        self.assertTrue(mock_validate.called)

    def test_update_wrong_key(self):
        body = {'quota_set': {'bad': 'bad'}}
        self.assertRaises(exception.InvalidInput, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_invalid_value_key_value(self):
        body = {'quota_set': {'gigabytes': "should_be_int"}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_invalid_type_key_value(self):
        body = {'quota_set': {'gigabytes': None}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_with_no_body(self):
        body = {}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_with_wrong_body(self):
        body = {'test': {}}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_multi_value_with_bad_data(self):
        orig_quota = self.controller.show(self.req, fake.PROJECT_ID)
        body = make_body(gigabytes=2000, snapshots=15, volumes="should_be_int",
                         backups=5, tenant_id=None)
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)
        # Verify that quota values are not updated in db
        new_quota = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(orig_quota, new_quota)

    def test_update_bad_quota_limit(self):
        body = {'quota_set': {'gigabytes': -1000}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)
        body = {'quota_set': {'gigabytes': db.MAX_INT + 1}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_no_admin(self):
        self.req.environ['cinder.context'].is_admin = False
        self.req.environ['cinder.context'].project_id = fake.PROJECT_ID
        self.req.environ['cinder.context'].user_id = 'foo_user'
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller.update, self.req, fake.PROJECT_ID,
                          body=make_body(tenant_id=None))

    def test_update_without_quota_set_field(self):
        body = {'fake_quota_set': {'gigabytes': 100}}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_update_empty_body(self):
        body = {}
        self.assertRaises(exception.ValidationError, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def _commit_quota_reservation(self):
        # Create simple quota and quota usage.
        ctxt = context.get_admin_context()
        res = test_db_api._quota_reserve(ctxt, fake.PROJECT_ID)
        db.reservation_commit(ctxt, res, fake.PROJECT_ID)
        expected = {'project_id': fake.PROJECT_ID,
                    'volumes': {'reserved': 0, 'in_use': 1},
                    'gigabytes': {'reserved': 0, 'in_use': 2},
                    }
        self.assertEqual(expected,
                         db.quota_usage_get_all_by_project(ctxt,
                                                           fake.PROJECT_ID))

    def test_update_lower_than_existing_resources(self):
        self._commit_quota_reservation()
        body = {'quota_set': {'volumes': 0}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)
        # Ensure that validation works even if some resources are valid
        body = {'quota_set': {'gigabytes': 1, 'volumes': 10}}
        self.assertRaises(webob.exc.HTTPBadRequest, self.controller.update,
                          self.req, fake.PROJECT_ID, body=body)

    def test_delete(self):
        result_show = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(make_body(), result_show)

        body = make_body(gigabytes=2000, snapshots=15,
                         volumes=5, backups=5,
                         backup_gigabytes=1000, tenant_id=None)
        result_update = self.controller.update(self.req, fake.PROJECT_ID,
                                               body=body)
        self.assertDictEqual(body, result_update)

        self.controller.delete(self.req, fake.PROJECT_ID)

        result_show_after = self.controller.show(self.req, fake.PROJECT_ID)
        self.assertDictEqual(result_show, result_show_after)

    def test_delete_with_allocated_quota_different_from_zero(self):
        project_id_1 = uuid.uuid4().hex
        project_id_2 = uuid.uuid4().hex
        self.req.environ['cinder.context'].project_id = project_id_1

        body = make_body(gigabytes=2000, snapshots=15,
                         volumes=5, backups=5,
                         backup_gigabytes=1000, tenant_id=None)
        result_update = self.controller.update(self.req, project_id_1,
                                               body=body)
        self.assertDictEqual(body, result_update)

        # Set usage param to True in order to see get allocated values.
        self.req.params = {'usage': 'True'}
        result_show = self.controller.show(self.req, project_id_1)

        result_update = self.controller.update(self.req, project_id_2,
                                               body=body)
        self.assertDictEqual(body, result_update)

        self.controller.delete(self.req, project_id_2)

        result_show_after = self.controller.show(self.req, project_id_1)
        self.assertDictEqual(result_show, result_show_after)

    def test_delete_no_admin(self):
        self.req.environ['cinder.context'].is_admin = False
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller.delete, self.req, fake.PROJECT_ID)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from http import HTTPStatus
import inspect
from unittest import mock

import ddt
from oslo_utils import encodeutils
import webob

from cinder.api.openstack import wsgi
from cinder import exception
from cinder.tests.unit.api import fakes
from cinder.tests.unit import test


class RequestTest(test.TestCase):
    def setUp(self):
        super(RequestTest, self).setUp()
        self.patch('cinder.i18n.get_available_languages',
                   return_value=['en-GB', 'en-AU', 'de', 'zh-CN', 'en-US',
                                 'ja-JP'])

    def test_content_type_missing(self):
        request = wsgi.Request.blank('/tests/123', method='POST')
        request.body = b"<body />"
        self.assertIsNone(request.get_content_type())

    def test_content_type_unsupported(self):
        request = wsgi.Request.blank('/tests/123', method='POST')
        request.headers["Content-Type"] = "text/html"
        request.body = b"asdf<br />"
        self.assertRaises(exception.InvalidContentType,
                          request.get_content_type)

    def test_content_type_with_charset(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Content-Type"] = "application/json; charset=UTF-8"
        result = request.get_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_from_accept(self):
        for content_type in ('application/json',
                             'application/vnd.openstack.volume+json'):
            request = wsgi.Request.blank('/tests/123')
            request.headers["Accept"] = content_type
            result = request.best_match_content_type()
            self.assertEqual(content_type, result)

    def test_content_type_from_accept_best(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Accept"] = "application/json"
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_from_query_extension(self):
        request = wsgi.Request.blank('/tests/123.json')
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

        request = wsgi.Request.blank('/tests/123.invalid')
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_accept_default(self):
        request = wsgi.Request.blank('/tests/123.unsupported')
        request.headers["Accept"] = "application/unsupported1"
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_accept_with_quality_values(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Accept"] = (
            "application/json;q=0.4,"
            "application/vnd.openstack.volume+json;q=0.6")
        result = request.best_match_content_type()
        self.assertEqual("application/vnd.openstack.volume+json", result)

    def test_from_request(self):
        request = wsgi.Request.blank('/')
        accepted = 'bogus;q=1, en-gb;q=0.7,en-us,en;q=0.5,*;q=0.7'
        request.headers = {'Accept-Language': accepted}
        self.assertEqual(request.best_match_language(), 'en-US')

    def test_asterisk(self):
        # In the section 3.4 of RFC4647, it says as follows:
        # If the language range "*"(asterisk) is the only one
        # in the language priority list or if no other language range
        # follows, the default value is computed and returned.
        #
        # In this case, the default value 'None' is returned.
        request = wsgi.Request.blank('/')
        accepted = '*;q=0.5'
        request.headers = {'Accept-Language': accepted}
        self.assertIsNone(request.best_match_language())

    def test_asterisk_followed_by_other_language(self):
        request = wsgi.Request.blank('/')
        accepted = '*,ja-jp;q=0.5'
        request.headers = {'Accept-Language': accepted}
        self.assertEqual('ja-JP', request.best_match_language())

    def test_truncate(self):
        request = wsgi.Request.blank('/')
        accepted = 'de-CH'
        request.headers = {'Accept-Language': accepted}
        self.assertEqual('de', request.best_match_language())

    def test_secondary(self):
        request = wsgi.Request.blank('/')
        accepted = 'nn,en-gb;q=0.5'
        request.headers = {'Accept-Language': accepted}
        self.assertEqual('en-GB', request.best_match_language())

    def test_none_found(self):
        request = wsgi.Request.blank('/')
        accepted = 'nb-no'
        request.headers = {'Accept-Language': accepted}
        self.assertIsNone(request.best_match_language())

    def test_no_lang_header(self):
        request = wsgi.Request.blank('/')
        accepted = ''
        request.headers = {'Accept-Language': accepted}
        self.assertIsNone(request.best_match_language())

    def test_best_match_language(self):
        # Test that we are actually invoking language negotiation by webob
        request = wsgi.Request.blank('/')
        accepted = 'unknown-lang'
        request.headers = {'Accept-Language': accepted}

        self.mock_object(request.accept_language,
                         'best_match', return_value=None)

        self.assertIsNone(request.best_match_language())
        # If accept-language is not included or empty, match should be None
        request.headers = {'Accept-Language': ''}
        self.assertIsNone(request.best_match_language())
        request.headers.pop('Accept-Language')
        self.assertIsNone(request.best_match_language())

    def test_cache_and_retrieve_resources(self):
        request = wsgi.Request.blank('/foo')
        # Test that trying to retrieve a cached object on
        # an empty cache fails gracefully
        self.assertIsNone(request.cached_resource())
        self.assertIsNone(request.cached_resource_by_id('r-0'))

        resources = []
        for x in range(3):
            resources.append({'id': 'r-%s' % x})

        # Cache an empty list of resources using the default name
        request.cache_resource([])
        self.assertEqual({}, request.cached_resource())
        self.assertIsNone(request.cached_resource('r-0'))
        # Cache some resources
        request.cache_resource(resources[:2])
        # Cache  one resource
        request.cache_resource(resources[2])
        # Cache  a different resource name
        other_resource = {'id': 'o-0'}
        request.cache_resource(other_resource, name='other-resource')

        self.assertEqual(resources[0], request.cached_resource_by_id('r-0'))
        self.assertEqual(resources[1], request.cached_resource_by_id('r-1'))
        self.assertEqual(resources[2], request.cached_resource_by_id('r-2'))
        self.assertIsNone(request.cached_resource_by_id('r-3'))
        self.assertEqual({'r-0': resources[0],
                          'r-1': resources[1],
                          'r-2': resources[2]}, request.cached_resource())
        self.assertEqual(other_resource,
                         request.cached_resource_by_id('o-0',
                                                       name='other-resource'))

    def test_cache_and_retrieve_volumes(self):
        self._test_cache_and_retrieve_resources('volume')

    def test_cache_and_retrieve_volume_types(self):
        self._test_cache_and_retrieve_resources('volume_type')

    def test_cache_and_retrieve_snapshots(self):
        self._test_cache_and_retrieve_resources('snapshot')

    def test_cache_and_retrieve_backups(self):
        self._test_cache_and_retrieve_resources('backup')

    def _test_cache_and_retrieve_resources(self, resource_name):
        """Generic helper for cache tests."""
        cache_all_func = 'cache_db_%ss' % resource_name
        cache_one_func = 'cache_db_%s' % resource_name
        get_db_all_func = 'get_db_%ss' % resource_name
        get_db_one_func = 'get_db_%s' % resource_name

        r = wsgi.Request.blank('/foo')
        resources = []
        for x in range(3):
            resources.append({'id': 'id%s' % x})

        # Store 2
        getattr(r, cache_all_func)(resources[:2])
        # Store 1
        getattr(r, cache_one_func)(resources[2])

        self.assertEqual(resources[0], getattr(r, get_db_one_func)('id0'))
        self.assertEqual(resources[1], getattr(r, get_db_one_func)('id1'))
        self.assertEqual(resources[2], getattr(r, get_db_one_func)('id2'))
        self.assertIsNone(getattr(r, get_db_one_func)('id3'))
        self.assertEqual({'id0': resources[0],
                          'id1': resources[1],
                          'id2': resources[2]}, getattr(r, get_db_all_func)())


class ActionDispatcherTest(test.TestCase):
    def test_dispatch(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        self.assertEqual('pants', serializer.dispatch({}, action='create'))

    def test_dispatch_action_None(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        serializer.default = lambda x: 'trousers'
        self.assertEqual('trousers', serializer.dispatch({}, action=None))

    def test_dispatch_default(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        serializer.default = lambda x: 'trousers'
        self.assertEqual('trousers', serializer.dispatch({}, action='update'))


class DictSerializerTest(test.TestCase):
    def test_dispatch_default(self):
        serializer = wsgi.DictSerializer()
        self.assertEqual('', serializer.serialize({}, 'update'))


class JSONDictSerializerTest(test.TestCase):
    def test_json(self):
        input_dict = dict(servers=dict(a=(2, 3)))
        expected_json = b'{"servers":{"a":[2,3]}}'
        serializer = wsgi.JSONDictSerializer()
        result = serializer.serialize(input_dict)
        result = result.replace(b'\n', b'').replace(b' ', b'')
        self.assertEqual(expected_json, result)


class TextDeserializerTest(test.TestCase):
    def test_dispatch_default(self):
        deserializer = wsgi.TextDeserializer()
        self.assertEqual({}, deserializer.deserialize({}, 'update'))


class JSONDeserializerTest(test.TestCase):
    def test_json(self):
        data = """{"a": {
                "a1": "1",
                "a2": "2",
                "bs": ["1", "2", "3", {"c": {"c1": "1"}}],
                "d": {"e": "1"},
                "f": "1"}}"""
        as_dict = {
            'body': {
                'a': {
                    'a1': '1',
                    'a2': '2',
                    'bs': ['1', '2', '3', {'c': {'c1': '1'}}],
                    'd': {'e': '1'},
                    'f': '1',
                },
            },
        }
        deserializer = wsgi.JSONDeserializer()
        self.assertEqual(as_dict, deserializer.deserialize(data))


class ResourceTest(test.TestCase):
    def test_resource_call(self):
        class Controller(object):
            def index(self, req):
                return 'off'

        req = webob.Request.blank('/tests')
        app = fakes.TestRouter(Controller())
        response = req.get_response(app)
        self.assertEqual(b'off', response.body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_resource_not_authorized(self):
        class Controller(object):
            def index(self, req):
                raise exception.NotAuthorized()

        req = webob.Request.blank('/tests')
        app = fakes.TestRouter(Controller())
        response = req.get_response(app)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_dispatch(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        method, _extensions = resource.get_method(None, 'index', None, '')
        actual = resource.dispatch(method, None, {'pants': 'off'})
        expected = 'off'
        self.assertEqual(expected, actual)

    @mock.patch('oslo_utils.strutils.mask_password')
    def test_process_stack_non_ascii(self, masker):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def fooAction(self, req, id, body):
                return 'done'

        controller = Controller()
        resource = wsgi.Resource(controller)
        # The following body has a non-ascii chars
        serialized_body = '{"foo": {"nonascii": "\xe2\x80\x96\xe2\x88\xa5"}}'
        request = webob.Request.blank('/tests/fooAction')
        action_args = {'id': 12}
        # Now test _process_stack() mainline flow.
        # Without the fix to safe_decode the body in _process_stack(),
        # this test fails with:
        #     UnicodeDecodeError: 'ascii' codec can't decode byte 0xe2 in
        #                         position 22: ordinal not in range(128)
        response = resource._process_stack(request, 'fooAction', action_args,
                                           'application/json', serialized_body,
                                           'application/json')
        self.assertEqual('done', response)
        # The following check verifies that mask_password was called with
        # the decoded body.
        self.assertEqual(1, masker.call_count)
        decoded_body = encodeutils.safe_decode(
            serialized_body, errors='ignore')
        self.assertIn(decoded_body, masker.call_args[0][0])

    def test_get_method_undefined_controller_action(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(AttributeError, resource.get_method,
                          None, 'create', None, '')

    def test_get_method_action_json(self):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        method, _extensions = resource.get_method(None, 'action',
                                                  'application/json',
                                                  '{"fooAction": true}')
        self.assertEqual(controller._action_foo, method)

    def test_get_method_action_bad_body(self):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(exception.MalformedRequestBody, resource.get_method,
                          None, 'action', 'application/json', '{}')

    def test_get_method_unknown_controller_action(self):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(KeyError, resource.get_method,
                          None, 'action', 'application/json',
                          '{"barAction": true}')

    def test_get_action_args(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        env = {
            'wsgiorg.routing_args': [None, {
                'controller': None,
                'format': None,
                'action': 'update',
                'id': 12,
            }],
        }

        expected = {'action': 'update', 'id': 12}

        self.assertEqual(expected, resource.get_action_args(env))

    def test_get_body_bad_content(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.headers['Content-Type'] = 'application/none'
        request.body = b'foo'

        self.assertRaises(exception.InvalidContentType,
                          resource.get_body, request)

    def test_get_body_no_content_type(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.body = b'foo'

        content_type, body = resource.get_body(request)
        self.assertIsNone(content_type)
        self.assertEqual('', body)

    def test_get_body_no_content_body(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.headers['Content-Type'] = 'application/json'
        request.body = b''

        content_type, body = resource.get_body(request)
        self.assertIsNone(content_type)
        self.assertEqual('', body)

    def test_get_body(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.headers['Content-Type'] = 'application/json'
        request.body = b'foo'

        content_type, body = resource.get_body(request)
        self.assertEqual('application/json', content_type)
        self.assertEqual(b'foo', body)

    def test_deserialize_badtype(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(exception.InvalidContentType,
                          resource.deserialize,
                          controller.index, 'application/none', 'foo')

    def test_deserialize_default(self):
        class JSONDeserializer(object):
            def deserialize(self, body):
                return 'json'

        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller, json=JSONDeserializer)

        obj = resource.deserialize(controller.index, 'application/json', 'foo')
        self.assertEqual('json', obj)

    def test_register_actions(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

            @wsgi.action('barAction')
            def _action_bar(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertEqual({}, resource.wsgi_actions)

        extended = ControllerExtended()
        resource.register_actions(extended)
        self.assertEqual({'fooAction': extended._action_foo,
                          'barAction': extended._action_bar, },
                         resource.wsgi_actions)

    def test_register_extensions(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.extends
            def index(self, req, resp_obj, pants=None):
                return None

            @wsgi.extends(action='fooAction')
            def _action_foo(self, req, resp, id, body):
                return None

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertEqual({}, resource.wsgi_extensions)
        self.assertEqual({}, resource.wsgi_action_extensions)

        extended = ControllerExtended()
        resource.register_extensions(extended)
        self.assertEqual({'index': [extended.index]}, resource.wsgi_extensions)
        self.assertEqual({'fooAction': [extended._action_foo]},
                         resource.wsgi_action_extensions)

    def test_get_method_extensions(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.extends
            def index(self, req, resp_obj, pants=None):
                return None

        controller = Controller()
        extended = ControllerExtended()
        resource = wsgi.Resource(controller)
        resource.register_extensions(extended)
        method, extensions = resource.get_method(None, 'index', None, '')
        self.assertEqual(controller.index, method)
        self.assertEqual([extended.index], extensions)

    def test_get_method_action_extensions(self):
        class Controller(wsgi.Controller):
            def index(self, req, pants=None):
                return pants

            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        class ControllerExtended(wsgi.Controller):
            @wsgi.extends(action='fooAction')
            def _action_foo(self, req, resp_obj, id, body):
                return None

        controller = Controller()
        extended = ControllerExtended()
        resource = wsgi.Resource(controller)
        resource.register_extensions(extended)
        method, extensions = resource.get_method(None, 'action',
                                                 'application/json',
                                                 '{"fooAction": true}')
        self.assertEqual(controller._action_foo, method)
        self.assertEqual([extended._action_foo], extensions)

    def test_get_method_action_whitelist_extensions(self):
        class Controller(wsgi.Controller):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.action('create')
            def _create(self, req, body):
                pass

            @wsgi.action('delete')
            def _delete(self, req, id):
                pass

        controller = Controller()
        extended = ControllerExtended()
        resource = wsgi.Resource(controller)
        resource.register_actions(extended)

        method, extensions = resource.get_method(None, 'create',
                                                 'application/json',
                                                 '{"create": true}')
        self.assertEqual(extended._create, method)
        self.assertEqual([], extensions)

        method, extensions = resource.get_method(None, 'delete', None, None)
        self.assertEqual(extended._delete, method)
        self.assertEqual([], extensions)

    def test_pre_process_extensions_regular(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return None

        def extension2(req, resp_obj):
            called.append(2)
            return None

        extensions = [extension1, extension2]
        response, post = resource.pre_process_extensions(extensions, None, {})
        self.assertEqual([], called)
        self.assertIsNone(response)
        self.assertEqual([extension2, extension1], list(post))

    def test_pre_process_extensions_generator(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            called.append('pre1')
            yield
            called.append('post1')

        def extension2(req):
            called.append('pre2')
            yield
            called.append('post2')

        extensions = [extension1, extension2]
        response, post = resource.pre_process_extensions(extensions, None, {})
        post = list(post)
        self.assertEqual(['pre1', 'pre2'], called)
        self.assertIsNone(response)
        self.assertEqual(2, len(post))
        self.assertTrue(inspect.isgenerator(post[0]))
        self.assertTrue(inspect.isgenerator(post[1]))

        for gen in post:
            try:
                gen.send(None)
            except StopIteration:
                continue

        self.assertEqual(['pre1', 'pre2', 'post2', 'post1'], called)

    def test_pre_process_extensions_generator_response(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            called.append('pre1')
            yield 'foo'

        def extension2(req):
            called.append('pre2')

        extensions = [extension1, extension2]
        response, post = resource.pre_process_extensions(extensions, None, {})
        self.assertEqual(['pre1'], called)
        self.assertEqual('foo', response)
        self.assertEqual([], post)

    def test_post_process_extensions_regular(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return None

        def extension2(req, resp_obj):
            called.append(2)
            return None

        response = resource.post_process_extensions([extension2, extension1],
                                                    None, None, {})
        self.assertEqual([2, 1], called)
        self.assertIsNone(response)

    def test_post_process_extensions_regular_response(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return None

        def extension2(req, resp_obj):
            called.append(2)
            return 'foo'

        response = resource.post_process_extensions([extension2, extension1],
                                                    None, None, {})
        self.assertEqual([2], called)
        self.assertEqual('foo', response)

    def test_post_process_extensions_version_not_found(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return 'bar'

        def extension2(req, resp_obj):
            raise exception.VersionNotFoundForAPIMethod(version='fake_version')

        response = resource.post_process_extensions([extension2, extension1],
                                                    None, None, {})
        self.assertEqual([1], called)
        self.assertEqual('bar', response)

    def test_post_process_extensions_generator(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            yield
            called.append(1)

        def extension2(req):
            yield
            called.append(2)

        ext1 = extension1(None)
        next(ext1)
        ext2 = extension2(None)
        next(ext2)

        response = resource.post_process_extensions([ext2, ext1],
                                                    None, None, {})

        self.assertEqual([2, 1], called)
        self.assertIsNone(response)

    def test_post_process_extensions_generator_response(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            yield
            called.append(1)

        def extension2(req):
            yield
            called.append(2)
            yield 'foo'

        ext1 = extension1(None)
        next(ext1)
        ext2 = extension2(None)
        next(ext2)

        response = resource.post_process_extensions([ext2, ext1],
                                                    None, None, {})

        self.assertEqual([2], called)
        self.assertEqual('foo', response)


class ResponseObjectTest(test.TestCase):
    def test_default_code(self):
        robj = wsgi.ResponseObject({})
        self.assertEqual(HTTPStatus.OK, robj.code)

    def test_modified_code(self):
        robj = wsgi.ResponseObject({})
        robj._default_code = HTTPStatus.ACCEPTED
        self.assertEqual(HTTPStatus.ACCEPTED, robj.code)

    def test_override_default_code(self):
        robj = wsgi.ResponseObject({}, code=HTTPStatus.NOT_FOUND)
        self.assertEqual(HTTPStatus.NOT_FOUND, robj.code)

    def test_override_modified_code(self):
        robj = wsgi.ResponseObject({}, code=HTTPStatus.NOT_FOUND)
        robj._default_code = HTTPStatus.ACCEPTED
        self.assertEqual(HTTPStatus.NOT_FOUND, robj.code)

    def test_set_header(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        self.assertEqual({'header': 'foo'}, robj.headers)

    def test_get_header(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        self.assertEqual('foo', robj['hEADER'])

    def test_del_header(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        del robj['hEADER']
        self.assertNotIn('header', robj.headers)

    def test_header_isolation(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        hdrs = robj.headers
        hdrs['hEADER'] = 'bar'
        self.assertEqual('foo', robj['hEADER'])

    def test_default_serializers(self):
        robj = wsgi.ResponseObject({})
        self.assertEqual({}, robj.serializers)


@ddt.data
class ValidBodyTest(test.TestCase):

    def setUp(self):
        super(ValidBodyTest, self).setUp()
        self.controller = wsgi.Controller()

    def test_assert_valid_body(self):
        body = {'foo': {}}
        self.controller.assert_valid_body(body, 'foo')

    def test_assert_valid_body_none(self):
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.assert_valid_body(None, 'foo'))

    def test_assert_valid_body_empty(self):
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.assert_valid_body({}, 'foo'))

    def test_assert_valid_body_no_entity(self):
        body = {'bar': {}}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.assert_valid_body(body, 'foo'))

    def test_assert_valid_body_malformed_entity(self):
        body = {'foo': 'bar'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.assert_valid_body(body, 'foo'))

    def test_validate_string_length_with_name_too_long(self):
        name = 'a' * 256
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.validate_string_length,
                          name, 'Name', min_length=1, max_length=255,
                          remove_whitespaces=False)

    @ddt.data('name', 'display_name', 'description', 'display_description')
    def test_validate_name_and_description_with_name_too_long(self, attribute):
        body = {attribute: 'a' * 256}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.validate_name_and_description,
                          body)

    @ddt.data('name', 'display_name', 'description', 'display_description')
    def test_validate_name_and_description_with_name_as_int(self, attribute):
        body = {attribute: 1234}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.validate_name_and_description,
                          body)

    @ddt.data('name', 'display_name', 'description', 'display_description')
    def test_validate_name_and_description_with_name_zero_length(self,
                                                                 attribute):
        # NOTE(jdg): We allow zero length names currently, particularly
        # from Nova, changes to this require an API version bump
        body = {attribute: ""}
        self.controller.validate_name_and_description(body)
        self.assertEqual('', body[attribute])

    @ddt.data('name', 'display_name', 'description', 'display_description')
    def test_validate_name_and_description_with_name_contains_white_spaces(
            self, attribute):
        body = {attribute: 'a' * 255 + "  "}
        self.controller.validate_name_and_description(body)
        self.assertEqual('a' * 255, body[attribute])



# Stubs for oslo_i18n (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from ._factory import *
from ._gettextutils import *
from ._lazy import *
from ._translate import *
